from tko.play.keys import GuiActions
from tko.settings.settings import Settings
from tko.util.text import Text
from tko.util.symbols import symbols
from pathlib import Path

from tko.play.fmt import Fmt
from tko.play.frame import Frame
from tko.play.border import Border
from tko.play.search import Search
from tko.play.language_setter import LanguageSetter
from tko.play.images import opening
from tko.play.floating import Floating
from tko.play.floating_manager import FloatingManager
from tko.play.flags import Flags, FlagsMan
from tko.play.tasktree import TaskTree, TaskAction

from tko.play.keys import GuiKeys
from tko.game.quest import Quest
from tko.game.task import Task
from tko.play.task_graph import TaskGraph
from tko.play.daily_graph import DailyGraph

class Gui:

    def __init__(self, tree: TaskTree, flagsman: FlagsMan, fman: FloatingManager):
        self.rep = tree.rep
        self.game = tree.game
        self.tree = tree
        self.flagsman = flagsman
        self.fman = fman
        self.settings = tree.settings
        self.search = Search(tree=self.tree, fman=self.fman)
        self.style: Border = Border(self.settings.app)
        self.language = LanguageSetter(self.settings, self.rep, self.flagsman, self.fman)
        self.colors = self.settings.colors
        self.need_update = False
        self.xray_offset = 0

        self.app = Settings().app

    def set_need_update(self):
        self.need_update = True

    def get_help_fixed(self):
        color = "W"
        try:
            selected = self.tree.get_selected_throw()
            if isinstance(selected, Task):
                color = "G"
        except IndexError:
            pass
        act_color, act_text = self.get_activate_label()
        help_fixed: list[Text] = [
            Text().addf("C", f"Criar Rascunho [{GuiKeys.create_draft}]"),
            Text().addf("C", f"{GuiActions.pallete} [{GuiKeys.palette}]"),
            Text().addf("C", f"{GuiActions.calibrate} [{GuiKeys.calibrate}]"),
            Text().addf("G", f"{GuiActions.search} [{GuiKeys.search}]"),
            Text().addf(act_color, f"{act_text} [↲]"),
            Text().addf(color, f"{GuiActions.grade} [{GuiKeys.self_evaluate}]"),
        ]
        return help_fixed

    def get_activate_label(self) -> tuple[str, str]:
        try:
            obj = self.tree.get_selected_throw()
        except IndexError:
            return "R", " Retornar"
        if isinstance(obj, Quest):
            quest: Quest = obj
            if Flags.inbox.get_value() != Flags.inbox_all and not quest.is_reachable():
                output = TaskAction.BLOQUEIO
            elif quest.get_full_key() in self.tree.expanded:
                output = TaskAction.CONTRAIR
            else:
                output = TaskAction.EXPANDIR
            return "Y", output
        elif isinstance(obj, Task):
            color, output = self.tree.get_task_action(obj)
            return color, output
        return "R", " ERRO"

    @staticmethod
    def get_admin_color() -> str:
        if Flags.inbox.get_value() == Flags.inbox_only:
            return "g"
        if Flags.inbox.get_value() == Flags.inbox_all:
            return "y"
        return "c"

    @staticmethod
    def center_header_footer(value: Text, frame: Frame) -> Text:
        half = value.len() // 2
        x = frame.get_x()
        _, dx = Fmt.get_size()
        color = Gui.get_admin_color()
        full = Text().addf(color, "─" * ((dx//2) - x - 2 - half)).add(value)
        return full

    def show_main_bar(self, frame: Frame):
        top = Text().add(" ")

        inbox: str = f"Inbox [{GuiKeys.inbox}]"
        left, right = self.get_lr(Flags.inbox.get_value() == Flags.inbox_only)
        top.add(left).addf("/", inbox).add(right).add(" ")

        all_tasks: str = f"Todas [{GuiKeys.all_tasks}]"
        left, right = self.get_lr(Flags.inbox.get_value() == Flags.inbox_all)
        top.add(left).addf("/", all_tasks).add(right).add(" ")

        toggle_right: str = f"Painel [{GuiKeys.toggle_right}]"
        left, right = self.get_lr(Flags.show_panel.get_value() == "1")
        top.add(left).addf("/", toggle_right).add(right).add(" ")

        helpv: str = f"Ajuda [{GuiKeys.panel_help}]"
        left, right = self.get_lr(Flags.panel.get_value() == Flags.panel_help and Flags.show_panel.get_value() == "1")
        top.add(left).addf("/", helpv).add(right).add(" ")
        
        # full = self.center_header_footer(top, frame)
        frame.set_header(top, "<")
        # footer            
        text = Text()
        alias_color = "R"
        dirname: Path = self.rep.paths.get_workspace_dir()
        dirname_str = dirname.name.upper()
        text.add(self.style.border(alias_color, dirname_str))
        text.add(self.style.border("G", self.rep.data.lang.upper()))
        if self.need_update:
            text = Text().addf("r", " TKO DESATUALIZADO!").addf("y"," Atualize com: ").addf("g", "pipx upgrade tko ")
        # text = self.center_header_footer(text, frame)
        frame.set_footer(text, "^")
        
        # if Flags.flags:
        #     value = self.make_flags_bar()
        #     full = self.center_header_footer(value, frame)
        #     frame.set_footer(full, "<")
        frame.draw()

        dy, dx = frame.get_inner()
        values = [Text()] + self.tree.get_senteces(dy)
        for y, sentence in enumerate(values):
            if sentence.len() > dx:
                sentence.trim_end(dx - 3)
                sentence.addf("r", "...")
            frame.write(y, 0, sentence)

    def show_skills_bar(self, frame_xp: Frame):
        dy, dx = frame_xp.get_inner()
        #_ = [q for q in self.game.quests.values() if q.is_reachable()]
        obtained, priority, complete = self.game.get_skills_resume()

        frame_xp.set_header(Text().add(self.gen_right_header()), ">")
        frame_xp.draw()

        elements: list[Text] = []
        for skill, value in complete.items():
            if Flags.show_panel.get_value() == "1":
                obtained_value = round(100 * obtained.get(skill, 0) / priority.get(skill, 1))
                possible_value = round(100 * value / priority.get(skill, 1))
                text = f"{skill}: {obtained_value:03d}%  {possible_value:03d}%"
            else:
                obtained_value = round(obtained.get(skill, 0))
                priority_value = round(priority.get(skill, 0))
                complete_value = round(value)
                text = f"{skill}:{obtained_value:03d}/{priority_value:03d}/{complete_value:03d}"
                # text = f"{skill}:{round(obtained.get(skill, 0))}/{round(value)}"
            perc = obtained.get(skill, 0) / priority.get(skill, 1)
            done_color = self.colors.progress_skill_done
            todo_color = self.colors.progress_skill_todo
            #text = text.rjust(dx - 4)
            skill_bar = self.style.build_bar(text=text, percent=perc, length=dx - 2, fmt_true=done_color, fmt_false=todo_color)
            elements.append(skill_bar)
        
        # Total bar
        total_obtained, total_priority, total_complete = self.game.sum_xp(obtained, priority, complete)
        if total_priority == 0:
            grade = 0
        else:
            grade = total_obtained / total_priority * 10.0

        if Flags.show_panel.get_value() == "1":
            text = f" Nota: {grade:.1f}       "
        else:
            text = f"Nota: {grade:.1f} :{round(total_obtained):03d}/{round(total_priority):03d}/{round(total_complete):03d}"
        # text = text.rjust(dx - 4)
        done_color = self.colors.main_bar_done
        todo_color = self.colors.main_bar_todo
        percent = total_obtained / total_priority if total_priority > 0 else 0.0
        total_bar = self.style.build_bar(text, percent , dx - 2, done_color, todo_color)
        elements.append(total_bar)

        # printing calculating line breaks
        line_breaks = dy - len(elements)
        for skill_bar in elements:
            frame_xp.print(1, skill_bar)
            if line_breaks > 0:
                line_breaks -= 1
                frame_xp.print(1, Text())

    def build_list_sentence(self, items: list[Text]) -> list[Text]:
        out: list[Text] = []
        for x in items:
            color_ini = x.data[0].fmt
            color_end = x.data[-1].fmt
            left = self.style.round_l(color_ini)
            right = self.style.round_r(color_end)
            middle = x.clone()
            if x.data[0].text == "!":
                left = self.style.sharp_l(color_ini)
                right = self.style.sharp_r(color_end)
                middle.data = x.data[1:]
            out.append(Text().add(left).add(middle).add(right))
        return out

    def build_bottom_array(self):
        array: list[Text] = []
        array += self.get_help_fixed()
        return self.build_list_sentence(array)

    def show_bottom_bar(self):
        lines, cols = Fmt.get_size()
        elems = self.build_bottom_array()
        line_main = Text().add(" ").join(elems) # alignment adjust
        Fmt.write(lines - 1, 0, line_main.center(cols))

    def make_xp_button(self, size: int):
        if self.search.search_mode:
            text = " Busca: " + self.tree.search_text + symbols.cursor.text
            percent = 0.0
            done_color = "W"
            todo_color = "W"
            text = text.ljust(size)
            return self.style.build_bar(text, percent, len(text), done_color, todo_color)
            # text, percent = self.build_xp_bar()
            # done_color = self.colors.main_bar_done
            # todo_color = self.colors.main_bar_todo
            # text = text.center(size)
            # xpbar = self.style.build_bar(text, percent, len(text), done_color, todo_color)

        #reachable_quests = [q for q in self.game.quests.values() if q.is_reachable()]
        obtained, priority, complete = self.game.get_skills_resume()

        keys_to_remove: list[str] = []
        for skill, value in obtained.items():
            if value < 1:
                keys_to_remove.append(skill)
        for key in keys_to_remove:
            if key in obtained:
                del obtained[key]
            if key in priority:
                del priority[key]
            if key in complete:
                del complete[key]

        qtd = len(obtained)
        if qtd == 0:
            text = " Nenhuma habilidade disponível "
            percent = 0.0
            return self.style.build_bar(text, percent, size, "W", "W")
        
        skill_size = int(size / qtd)

        elements: list[Text] = []
        for skill, _ in complete.items():
            text = f"{skill}"
            perc = obtained.get(skill, 0) / priority.get(skill, 1)
            done_color = self.colors.main_bar_done
            todo_color = self.colors.main_bar_todo
            # text = text.rjust(skill_size - 4)
            skill_bar = self.style.build_bar(text=text, percent=perc, length=skill_size - 2, fmt_true=done_color, fmt_false=todo_color, rounded=False)
            elements.append(skill_bar)
        cover_color = 'K'
        xpbar = Text().add(self.style.round_l(cover_color)).addf(cover_color.lower(), '█')
        for skill_bar in elements:
            xpbar.add(skill_bar).addf(cover_color.lower(), '█')
        xpbar.add(self.style.round_r(cover_color))

        return xpbar

    def show_top_bar(self, frame: Frame):
        lista = [
            Text().addf('Y', f"Sair [esc]"),
            Text().addf("Y", f"Ajuda[{GuiKeys.panel_help}]"),
        ]
        help_list = self.build_list_sentence(lista)
        pre: list[Text] = [help_list[0]]
        pos: list[Text] = [help_list[1]]

        limit = frame.get_dx()
        size = limit - Text().add(" ").join(pre + pos).len() - 2
        main_label = self.make_xp_button(size)
        info = Text().add(" ").join(pre + [main_label] + pos)
        frame.write(0, 0, info.center(frame.get_dx()))

    def show_help(self, frame: Frame):
        # def empty(value):
        #     pass

        frame.set_header(self.gen_right_header(), ">")
        frame.draw()

        help_lines: list[Text] = []

        help_lines.append(Text().add(""))
        help_lines.append(Text.format("    Ajuda ").addf("r", GuiKeys.panel_help).add("  Abre essa tela de ajuda")
        )

        help_lines.append(Text.format("  ").addf("r", "Shift + B")
                           .add("  Habilita ").addf("r", "").addf("R", "ícones").addf("r", "").add(" se seu ambiente suportar"))
        help_lines.append(Text() + "  " + Text.r_token("g", f"Shift + {GuiKeys.calibrate}") + "  Para calibrar os direcionais do teclado")
        help_lines.append(Text() + "  " + Text.r_token("g", " setas") + "     Para navegar entre os elementos")
    
        help_lines.append(Text() + f"   {GuiActions.pallete} " + Text.r_token("r", f"{GuiKeys.palette}") + "  Abre o menu de ações e configurações")
        help_lines.append(Text() + f"   {GuiActions.download} " + Text.r_token("r", f"{GuiKeys.down_task}") + "  Tenta baixar novamente a tarefa selecionada")
        help_lines.append(Text() + f"   {GuiActions.edit} " + Text.r_token("r", f"{GuiKeys.edit}") + "  Abre os arquivos no editor de código")
        help_lines.append(Text() + f"   {GuiActions.show_time} " + Text.r_token("r", f"{GuiKeys.show_duration}") + "  Alterna entre mostrar o tempo gasto nas tarefas ou não")
        help_lines.append(Text() + f"   Apagar " + Text.r_token("r", f"{GuiKeys.delete_folder}") + "  Apaga a pasta da tarefa selecionada (útil para resetar o progresso de uma tarefa)")
        help_lines.append(Text() + f"   {GuiActions.activate} " + Text.r_token("r", "↲") + "  Interage com o elemento de acordo com o contexto")
        help_lines.append(Text() + "             (baixar, visitar, escolher, compactar, expandir)")
        help_lines.append(Text() + f"  {GuiActions.grade} " + Text.r_token("r", GuiKeys.self_evaluate) + "  Abre tela para auto avaliação")
        help_lines.append(Text() + f"    {GuiActions.search} " + Text.r_token("r", f"{GuiKeys.search}") + "  Abre a barra de pesquisa")
        help_lines.append(Text() + f"  {GuiActions.draft} " + Text.r_token("r", f"{GuiKeys.create_draft}") + "  Cria um rascunho para escrever código ou anotações")

        help_lines.append(Text())
        help_lines.append(Text() + "Você pode mudar o editor padrão com o comando no terminal")
        help_lines.append(Text() + Text.r_token("g", "             tko config --editor <comando>"))
        help_lines.append(Text())

        for i, line in enumerate(help_lines):
            frame.write(i, 0, line)

    def get_task_graph(self, task_key: str, width: int, height: int) -> tuple[bool, list[Text], list[Text]]:
        tg = TaskGraph(self.settings, self.rep, task_key, width, height)
        header, graph = tg.get_output()
        if len(graph) == 0:
            return False, [], []
        # for y, line in enumerate(graph):
        #     frame.write(y, x, Text().addf("g", line))
        return True, header, graph

    def get_daily_graph(self, width: int, height: int) -> tuple[bool, list[Text]]:
        graph = DailyGraph(self.rep.logger, width, height).get_graph()
        if len(graph) == 0:
            return False, []
        # for y, line in enumerate(graph):
        #     frame.write(y, x, Text().addf("g", line))
        return True, graph

    def get_lr(self, test: bool) -> tuple[str, str]:
        if test:
            return "▶", "◀"
        else:
            return " ", " "

    def gen_right_header(self) -> Text:
        top = Text().add(" ")

        graph: str = f"Gráficos [{GuiKeys.panel_graph}]"
        left, right = self.get_lr(Flags.panel.get_value() == Flags.panel_graph)
        top.add(left).addf("/", graph).add(right).add(" ")

        logs: str = f"Logs [{GuiKeys.panel_logs}]"
        left, right = self.get_lr(Flags.panel.get_value() == Flags.panel_logs)
        top.add(left).addf("/", logs).add(right).add(" ")

        skills: str = f"Trilhas [{GuiKeys.panel_skills}]"
        left, right = self.get_lr(Flags.panel.get_value() == Flags.panel_skills)
        top.add(left).addf("/", skills).add(right).add(" ")
        return top

    def show_graphs(self, frame: Frame):
        
        try:
            selected = self.tree.get_selected_throw()
        except IndexError:
            selected = None

        lines, cols = frame.get_inner()
        
        # now: datetime.datetime = datetime.datetime.now()
        # parrot = random_get(opening, str(now.hour))
        keep_borders = False
        made = False
        list_data: list[Text] = []
        width = cols - 2
        if width < 5:
            width = 5
        header: list[Text] = []
        if Flags.panel.get_value() in (Flags.panel_graph, Flags.panel_logs):
            height = lines - 4
            if height < 3:
                height = 3
            if isinstance(selected, Task):
                made, header, list_data = self.get_task_graph(selected.get_full_key(), width, height)
            elif isinstance(selected, Quest):
                made, list_data = self.get_daily_graph(width, height)
        if not made:
            list_data = [Text().add(x).rjust(width) for x in opening["parrot"].splitlines()]
            keep_borders = True
        if self.xray_offset < 0:
            self.xray_offset = 0
        if self.xray_offset >= len(list_data):
            self.xray_offset = len(list_data) - 1
        
        offset = 0
        
        if Flags.panel.get_value() == Flags.panel_logs and isinstance(selected, Task):
            offset = self.xray_offset
            keep_borders = True
        count = -1
        line_count = 0
        
        if not keep_borders:
            frame.set_border_none()

        frame.set_header(self.gen_right_header(), ">")
        frame.draw()
        if header:
            for y, line in enumerate(header):
                frame.write(y, 0, line)
            line_count = len(header)
        for line in list_data:
            count += 1
            if count < offset:
                continue
            frame.write(line_count, 0, line)
            line_count += 1

    def show_items(self):
        border_color = self.get_admin_color()
        Fmt.clear()
        self.tree.reload_sentences()
        lines, cols = Fmt.get_size()
        main_sx = cols  # tamanho em x livre
        main_sy = lines  # size em y avaliable

        top_y = -1
        top_dy = 1  #quantas linhas o topo usa
        bottom_dy = 1 # quantas linhas o fundo usa
        mid_y = top_dy # onde o meio começa
        mid_sy = main_sy - top_dy - bottom_dy # tamanho do meio
        # left_size = 29

        right_sx = 0
        if Flags.show_panel.is_true():
            right_sx = cols - self.tree.get_total_width() - 1
            if right_sx > cols // 2:
                right_sx = cols // 2
        
        task_sx = main_sx - right_sx

        frame_top = Frame(top_y, 0).set_size(top_dy + 2, cols)
        self.show_top_bar(frame_top)

        # frame_bottom = Frame(lines - bottom_dy - 1, -1).set_size(bottom_dy + 2, cols + 2)
        self.show_bottom_bar()
        # if task_sx > 5: 
        frame_main = Frame(mid_y, 0).set_size(mid_sy, task_sx).set_border_color(border_color)
        self.show_main_bar(frame_main)
        # self.show_graphs(frame_main)

        if Flags.show_panel.is_true():
            frame_right = Frame(mid_y, cols - right_sx).set_size(mid_sy, right_sx).set_border_color(border_color)
            if Flags.panel.get_value() == Flags.panel_skills:
                self.show_skills_bar(frame_right)
            elif Flags.panel.get_value() in [Flags.panel_logs, Flags.panel_graph]:
                self.show_graphs(frame_right)
            elif Flags.panel.get_value() == Flags.panel_help:
                self.show_help(frame_right)



import argparse

from tko.feno.build import build_main
from tko.feno.indexer import indexer_main
from tko.feno.html import html_main
from tko.feno.mdpp import mdpp_main
from tko.feno.older import older_main
from tko.feno.remote_md import remote_main
from tko.feno.filter import filter_main
from tko.feno.grading import Grading


def feno_main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title='subcommands', help='Help for subcommand.')

    # subparser for the 'build' command
    parser_r = subparsers.add_parser('build', help='Build .mapi file.')
    parser_r.add_argument('targets', metavar='T', type=str, nargs='*', help='folders')
    parser_r.add_argument("--check", "-c", action="store_true", help="Check if the file needs to be rebuilt")
    parser_r.add_argument("--brief", "-b", action="store_true", help="Brief mode")
    # parser_r.add_argument("--pandoc", "-p", action="store_true", help="Use pandoc rather than python markdown")
    parser_r.add_argument("--remote", "-r", action="store_true", help="Search for remote.cfg file and create absolute links")
    parser_r.add_argument("--erase", "-e", action="store_true", help="Erase .md and .tio temp files")
    parser_r.add_argument("--debug", "-d", action='store_true', help="Display debug msgs")
    parser_r.set_defaults(func=build_main)

    # subparser for the 'indexer' command
    parser_i = subparsers.add_parser('indexer', help='Index Readme file.')
    parser_i.add_argument('path', type=str, help='Path to Markdown file')
    parser_i.add_argument("base", type=str, help="Folder with the problems")
    parser_i.set_defaults(func=indexer_main)

    # subparser for the 'html' command
    parser_h = subparsers.add_parser('html', help='Generate HTML file from markdown file.')
    parser_h.add_argument('input', type=str, help='Input markdown file')
    parser_h.add_argument('output', type=str, help='Output HTML file')
    parser_h.add_argument('--title', type=str, default="Problema", help='Title of the HTML file')
    parser_h.set_defaults(func=html_main)

    # subparser for the 'mdpp' command
    parser_m = subparsers.add_parser('mdpp', help='Preprocessor for markdown files.')
    parser_m.add_argument('targets', metavar='T', type=str, nargs='*', help='Readmes or folders')
    parser_m.add_argument('--quiet', '-q', action="store_true", help='quiet mode')
    parser_m.add_argument('--clean', '-c', action="store_true", help='clean mode')
    parser_m.set_defaults(func=mdpp_main)

    # subparser for the 'older' command
    parser_o = subparsers.add_parser('older', help='Check if the source is newer than the target')
    parser_o.add_argument('targets', type=str, nargs='+', help='Target files or directories')
    parser_o.set_defaults(func=older_main)

    # subparser for the 'remote' command
    parser_rm = subparsers.add_parser('remote', help='Create remote markdown file.')
    parser_rm.add_argument('target', type=str, help='Folders')
    parser_rm.add_argument('--output', '-o', type=str, help='Output file')
    parser_rm.set_defaults(func=remote_main)

    # subparser for the 'filter' command
    parser_f = subparsers.add_parser('filter', help='Filter code removing answers.')

    parser_f.add_argument('target', type=str, help='file or folder to process')
    parser_f.add_argument('-u', '--update', action="store_true", help='update source file')
    parser_f.add_argument('-c', '--cheat', action="store_true", help='recursive cheat mode cleaning comments on students files')
    parser_f.add_argument('-o', '--output', type=str, help='output target')
    parser_f.add_argument("-r", "--recursive", action="store_true", help="recursive mode")
    parser_f.add_argument("-f", "--force", action="store_true", help="force mode")
    parser_f.add_argument("-q", "--quiet", action="store_true", help="quiet mode")
    parser_f.add_argument("-i", "--indent", type=int, default=0, help="indent using spaces")
    parser_f.set_defaults(func=filter_main)

    parser_grading = subparsers.add_parser('grading', help='Grade the tests.')

    parser_grading.add_argument("--output", "-o", type=str, help="Output file for the awarded grade.")
    grading_exclusive = parser_grading.add_mutually_exclusive_group(required=True)
    grading_exclusive.add_argument("--config", "-c", type=str, help="Path to the configuration file.")
    grading_exclusive.add_argument("--readme", "-r", type=str, help="Path to the README file for questions")
    parser_grading.set_defaults(func=Grading.main)

    args = parser.parse_args()
    if "func" not in args:
        parser.print_help()
        exit(1)
    else:
        args.func(args)



class KV:
    @staticmethod
    def decode_line(line: str, sep: str) -> dict[str, str]:
        args: list[str] = line.split(sep)
        return KV.decode_args(args)

    @staticmethod
    def decode_args(args: list[str]) -> dict[str, str]:
        kv: dict[str, str] = {}
        for value in args:
            parts = value.split(":")
            if len(parts) == 2:
                key = parts[0].strip()
                value = parts[1].strip()
                kv[key] = value
        return kv
    
    @staticmethod
    def encode_kv(kv: dict[str, str], sep: str = ", ") -> str:
        items: list[str] = []
        for k, v in kv.items():
            items.append(f"{k}:{v}")
        return sep.join(items)



from typing import Callable
from tko.settings.app_settings import AppSettings
import curses
import os

FN_VOID = Callable[[], None]

class InputManager:
    a = 97
    d = 100
    tab = 9
    special_double_key = 195
    cedilha = 167
    plus = 465
    minus = 464
    delete = 330

    def __init__(self):
        # stores a function than can return another function
        self.calls: dict[int, Callable[[], FN_VOID] | Callable[[], None]] = {}

    def add_int(self, _key: int, fn: Callable[[], None]):
        if _key in self.calls.keys():
            raise ValueError(f"Chave duplicada {chr(_key)}")
        self.calls[_key] = fn

    def add_str(self, str_key: str, fn: Callable[[], None]):
        if str_key != "":
            self.add_int(ord(str_key), fn)

    def has_str_key(self, key: str) -> bool:
        return ord(key) in self.calls
    
    def has_int_key(self, key: int) -> bool:
        return key in self.calls
    
    def exec_call(self, key: int):
        return self.calls[key]()

    @staticmethod
    def fix_esc_delay():
        if hasattr(curses, "set_escdelay"):
            curses.set_escdelay(25)
        else:
            os.environ.setdefault('ESCDELAY', '25')

    @staticmethod
    def get_and_remap_keys(scr: curses.window, app: AppSettings) -> int:
        value = scr.getch()
        if value == InputManager.special_double_key:
            value = scr.getch()
            if value == InputManager.cedilha: #ç
                value = ord("c")
        elif value == app.key_down:
            value = curses.KEY_DOWN
        elif value == app.key_up:
            value = curses.KEY_UP
        elif value == app.key_left:
            value = curses.KEY_LEFT
        elif value == app.key_right:
            value = curses.KEY_RIGHT
        elif value == app.key_backspace:
            value = curses.KEY_BACKSPACE
        elif value in [app.key_esc, 27]: # 27 is the ESC key value in some systems
            value = curses.KEY_EXIT
        elif value == app.key_pg_up:
            value = curses.KEY_PPAGE
        elif value == app.key_pg_down:
            value = curses.KEY_NPAGE
        # elif value > ord('A') and value < ord('Z'):
        #     value += 32 # convert to lowercase
        return value


from tko.play.floating import Floating, FloatingABC
from tko.util.symbols import symbols
from tko.util.text import Text


import curses
from typing import Callable


class FloatingInputText(FloatingABC):
    def __init__(self, label: Text, action: Callable[[str], None], unallowed_strings: list[str]| None = None):
        self.floating = Floating()
        self.floating.set_footer(" Use Enter para confirmar ou Esc para cancelar ")
        self.floating.set_text_ljust()
        self.floating.frame.set_border_color("m")
        self.floating.set_header(" Entrada de texto ")
        self.label = label
        self.input_text: str = ""
        self.action = action
        self.floating.frame.set_border_color("y")
        if unallowed_strings is None:
            unallowed_strings = []
        self.unallowed_strings = unallowed_strings

    # @override
    def is_enable(self) -> bool:
        return self.floating.is_enable()

    # @override
    def draw(self):
        content = self.floating.content
        content.clear()
        content.append(self.label)
        is_allowed = Text().addf("g", symbols.check.text) if self.input_text not in self.unallowed_strings else Text().addf("r", symbols.failure.text)
        content.append(Text.format("Digite aqui: ") + self.input_text + symbols.cursor + is_allowed)
        content.append(Text())
        self.floating.draw()

    # @override
    def process_input(self, key: int) -> int:
        if key == ord('\n'):
            if self.input_text in self.unallowed_strings:
                return -1
            self.floating.enable = False
            self.action(self.input_text)
            return -1
        elif key == curses.KEY_EXIT:
            self.floating.enable = False
            return -1
        elif key == curses.KEY_BACKSPACE:
            self.input_text = self.input_text[:-1]
            return -1
        elif 32 <= key < 127:
            self.input_text += chr(key)
            return -1
        return key


from __future__ import annotations
from tko.game.task import Task
from tko.util.text import Text
from tko.game.tree_item import TreeItem
from tko.game.quest_grader import QuestGrader
from tko.play.flags import Flags

# from typing import override

def startswith(text: str, prefix: str) -> bool: 
    if len(prefix) > len(text):
        return False
    return text[:len(prefix)] == prefix

class Quest(TreeItem):
    def __init__(self, title: str = "", key: str = ""):
        super().__init__()
        self.set_key(key)
        self.set_title(title)
        self.line_number: int = 0
        self.line: str = ""
        self.__tasks: list[Task] = []
        self.requires: list[str] = []  # 
        self.requires_ptr: list[Quest] = []
        self.skills: dict[str, int] = {}  # s:{skill} to be applied to all tasks
        self.languages: list[str] = []  # l:language to filter what is showed to user based in default language
        self.min_percent_completion: int = 50  # q:{value} 50 percent to complete quest
        self.filename = ""
        self.remote_name = ""
        self.__is_reachable: bool = False

    def add_require_key(self, key: str):
        if key.startswith("@"):
            key = key[1:]
        self.requires.append(self.get_remote_name() + "@" + key)

    def get_full_title(self) -> Text:
        output = Text().add(self.remote_name).add(":").add(self.get_title())
        if Flags.show_panel.get_value() == "1" and Flags.panel.get_value() == Flags.panel_skills:
            for skill, value in self.skills.items():
                if value > 1:
                    output.addf('b', f" +{skill}*{value}")
                else:
                    output.addf('b', f" +{skill}")
        return output

    def is_reachable(self)-> bool:
        return self.__is_reachable

    def set_reachable(self, value: bool):
        self.__is_reachable = value
        return self
    
    def update_tasks_reachable(self):
        for t in self.__tasks:
            t.set_reachable(True)
        return

    # @override
    def __str__(self):
        line = str(self.line_number).rjust(3)
        tasks_size = str(len(self.__tasks)).rjust(2, "0")
        key = "" if self.get_full_key() == self.get_title() else self.get_full_key() + " "
        output = f"{line} {tasks_size} {key}{self.get_title()} {self.skills} {self.requires}"
        return output

    def get_resume_by_percent(self) -> Text:
        value: int = min(100, round(self.get_percent()))
        return Text().addf(self.get_grade_color(), (str(value) + "%").rjust(4))
    
    def get_requirement(self) -> Text:
        return Text().addf("y", f"[{self.min_percent_completion}%]")

    def get_grade_color(self) -> str:
        if self.not_started():
            return "m"
        if not self.is_complete():
            return "r"
        if self.get_percent() == 100:
            return "g"
        return "y"

    def is_complete(self):
        return self.get_percent() >= self.min_percent_completion

    def add_task(self, task: Task):
        task.skills.update(self.skills)  # apply quest skills to task
        self.__tasks.append(task)

    def get_tasks(self):
        return self.__tasks

    def get_xp(self) -> tuple[float, float]:
        """
        Returns a tuple of (earned_xp, total_xp)
        """
        tasks_info: list[QuestGrader.Elem] = []
        for t in self.__tasks:
            tasks_info.append(QuestGrader.Elem(t.is_optional(), t.xp, t.get_percent()))
        return QuestGrader.calc_xp_earned_total(tasks_info)
        
    def get_percent(self) -> float:
        obtained, total = self.get_xp()
        return QuestGrader.get_percent(obtained, total)

    def in_progress(self):
        if self.is_complete():
            return False
        for t in self.__tasks:
            if t.get_percent() != 0:
                return True
        return False

    def not_started(self):
        if self.is_complete():
            return False
        if self.in_progress():
            return False
        return True



from tko.play.floating import Floating
from tko.play.floating_manager import FloatingManager
from tko.util.text import Text
from tko.settings.settings import Settings
from tko.down.drafts import Drafts
from pathlib import Path

import tempfile
import os
import subprocess

class Opener:
    def __init__(self, settings: Settings):
        self.settings = settings
        self.fman: None | FloatingManager = None
        self.folders: list[Path] = []
        self.language: str = ""

    def set_fman(self, fman: FloatingManager):
        self.fman = fman
        return self
    
    def set_target(self, folders: list[Path]):
        self.folders = folders
        return self

    def set_language(self, language: str):
        self.language = language
        return self

    def open_files(self, files_to_open: list[str]):
        files_raw = list(set(files_to_open))
        files_patched: list[str] = []
        for f in files_raw:
            if " " in f and not f.startswith('"'):
                files_patched.append(f'"{f}"')
            else:
                files_patched.append(f)

        cmd = self.settings.app.get_editor()
        folder = os.path.dirname(os.path.abspath(files_to_open[0]))
        aviso = (Floating()
                .bottom()
                .right()
                .set_warning()
                .put_sentence(Text().add("Pasta: ").addf("g", folder).add(" "))
                .put_text("Abrindo arquivos com o comando")
                )
        files = [os.path.basename(path) for path in files_to_open]
        aviso.put_sentence(Text().addf("g", f"{cmd}").add(" ").addf("g", " ".join(files)).add(" "))
        self.send_floating(aviso)
        fullcmd = "{} {}".format(cmd, " ".join(files_patched))
        outfile = tempfile.NamedTemporaryFile(delete=False)
        subprocess.Popen(fullcmd, stdout=outfile, stderr=outfile, shell=True)

    def send_floating(self, floating: Floating):
        if self.fman is not None:
            self.fman.add_input(floating)

    @staticmethod
    def try_add(files_to_open: list[Path], folder: Path, file: Path):
        path = folder / file
        if path.is_file():
            files_to_open.append(path)

    def load_folder(self, folder: Path) -> list[Path]:
        files_to_open: list[Path] = []
        files_to_open += Drafts.load_drafts_only(folder, self.language, ["md"])
        return files_to_open

    def load_folders_and_open(self):
        files_to_open: list[Path] = []
        for folder in self.folders:
            files_to_open += self.load_folder(folder)
        if len(files_to_open) != 0:
            self.open_files([str(f) for f in files_to_open])

    def __call__(self):
        self.load_folders_and_open()


from tko.logger.log_item_base import LogItemBase
from tko.game.task_info import TaskInfo
from tko.game.task import Task
from tko.logger.kv import KV

class LogItemSelf(LogItemBase):
    def __init__(self):
        super().__init__(LogItemBase.Type.SELF)
        self.info: TaskInfo = TaskInfo()

    def set_task(self, task: Task):
        self.key = task.get_full_key()
        self.info = task.info.clone()
        return self

    def set_info(self, info: TaskInfo):
        self.info = info
        return self
    
    def get_info(self) -> TaskInfo:
        return self.info

    def identify_kv(self, kv: dict[str, str]) -> bool:
        self.info.load_from_kv(kv)
        if self.type != LogItemBase.Type.SELF:
            return False
        return True

    def encode_line(self) -> str:
        output = super().encode_line()
        kv = self.info.get_kv()
        if len(kv) > 0:
            output += ", " + KV.encode_kv(kv, ", ")
        return output



import unicodedata

def uni_to_asc(input_str: str):
    nfkd_form = unicodedata.normalize('NFKD', input_str)
    return u"".join([c for c in nfkd_form if not unicodedata.combining(c)])

class SearchAsc:
    def __init__(self, pattern: str):
        self.pattern = uni_to_asc(pattern.lower())

    def find(self, title: str) -> int:
        return uni_to_asc(title.lower()).find(self.pattern)
    
    def inside(self, title: str) -> bool:
        return self.find(title) != -1


from tko.settings.settings import Settings
from tko.settings.repository import Repository
from uniplot import plot_to_string # type: ignore
from tko.util.text import Text
from tko.logger.log_sort import LogSort
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_item_base import LogItemBase
from tko.util.miniwi import Miniwi
from tko.logger.delta import Delta
from tko.play.flags import Flags

class TaskGraph:
    def __init__(self, settings: Settings, rep: Repository, task_key: str, width: int, height: int):
        self.settings = settings
        self.rep = rep
        self.task_key = task_key
        self.width = width
        self.height = height - 1
        
        self.collected_rate: list[float] = []
        self.collected_elapsed: list[float] = []
        self.collected_lines_len: list[float] = []

        self.total_elapsed: float = 0.0
        self.max_lines: float = 0
        self.actual_rate: float = 0.0
        
        self.eixo: list[float] = []
        self.logger = rep.logger
        self.versions = 0

        self.log_sort: LogSort | None = None
        if task_key in self.logger.tasks.task_dict:
            self.log_sort = self.logger.tasks.task_dict[task_key]

        self.raw_text: list[Text] = self.prepare_xray()

        if self.log_sort is None:
            return
        self.versions = len(self.log_sort.diff_list)
       
        item_exec_list: list[tuple[Delta, LogItemExec]] = self.log_sort.exec_list
        collected_rate: list[float] = [0]
        collected_elapsed: list[float] = [0]
        collected_lines_len: list[float] = [0]
        last = 0
        eixo: list[float] = [0]
        count = 1
        last_size = 0
        for delta, item in item_exec_list:
            if item.rate == -1:
                collected_rate.append(last)
            else:
                last = item.rate
                collected_rate.append(last)
            collected_elapsed.append(delta.accumulated.total_seconds() / 60)  # Convert to minutes
            if item.size > 0:
                last_size = item.size
            collected_lines_len.append(last_size)
            eixo.append(count)
            count += 1
        self.actual_rate = collected_rate[-1] if collected_rate else 0

        self.total_elapsed = collected_elapsed[-1]
        # normalizando tempo para porcentagem
        if not Flags.task_graph_mode.get_value() == Flags.task_time_view:
            if collected_elapsed[-1] != 0:
                for i in range(len(collected_elapsed)):
                    collected_elapsed[i] = collected_elapsed[i] / self.total_elapsed * 100

        # normalizando linhas para porcentagem        
        self.max_lines = max(collected_lines_len)
        if self.max_lines != 0:
            for i in range(len(collected_lines_len)):
                collected_lines_len[i] = (collected_lines_len[i] / self.max_lines) * 100
                if collected_lines_len[i] < 1:
                    collected_lines_len[i] = 0

        self.collected_rate = collected_rate
        self.collected_elapsed = collected_elapsed
        self.collected_lines_len = collected_lines_len
        self.eixo = eixo
        # self.eixo = list(range(len(collected)))

    def prepare_xray(self) -> list[Text]:
        if self.log_sort is None:
            return []
        output: list[Text] = []
        all_entries: list[tuple[Delta, LogItemBase]] = self.log_sort.base_list
        for delta, item in all_entries:
            data = str(item)
            data = data.split(", ")
            data = [x for x in data if not x.startswith("k:") and not x.startswith("v:")]
            data_str = ", ".join(data)
            data_str = data_str.replace("mode:", "")
            text = (Text().add(data_str)
                        .replace("EXEC", Text.Token("EXEC", "g"))
                        .replace("SELF", Text.Token("SELF", "r"))
                        .replace("MOVE", Text.Token("MOVE", "y")))
            acc = delta.accumulated.total_seconds() / 60
            acc_h = int(acc) // 60
            acc_m = int(acc) % 60
            acc_s = int((acc - int(acc)) * 60)
            acc_str = f"{acc_h:02d}:{acc_m:02d}:{acc_s:02d}"
            output.append(Text().add(f"acc:{acc_str}, ").add(text))
        return output

    def get_graph(self) -> list[Text]:
        if Flags.task_graph_mode.get_value() == Flags.task_time_view:
            result = plot_to_string(
                color=["magenta", "green", "red"],
                xs=[self.collected_elapsed, self.collected_elapsed, self.collected_elapsed],
                ys=[self.collected_lines_len, self.collected_rate, self.collected_rate],
                lines=[True, True, False], 
                y_min=0,
                y_max=101,
                width=self.width,
                height=self.height,
                y_unit="%",
                x_unit="min"
            )
        else:
            result = plot_to_string(
                color=["magenta", "green", "blue", "red"], 
                xs=[self.eixo, self.eixo, self.eixo, self.eixo], 
                ys=[self.collected_lines_len, self.collected_rate, self.collected_elapsed, self.collected_rate], 
                lines=[True, True, True, False], 
                y_min=0, 
                y_max=101, 
                width=self.width, 
                height=self.height, 
                y_unit="%", 
                x_unit="runs"
            )
        
        if isinstance(result, str):
            result = result.splitlines()
        output: list[Text] = []
        for line in result:
            output.append(Text.decode_raw(line))
        return output

    # returns header and graph lines
    def get_output(self) -> tuple[list[Text], list[Text]]:
        if not self.eixo:
            return [], []
        
        _ = Miniwi.render_dotted(self.task_key)
        first, second = "", ""
        
        # title = Text.format(" {C}", f" {self.task_key} ")
        title = Text()
        title += Text.format(" {G}", f" Total {self.actual_rate:.0f}% ")
        time_h: int = int(self.total_elapsed) // 60
        time_m: int = (int(self.total_elapsed) % 60)
        time = f"{time_h:02.0f}h {time_m:.0f}min" if time_h > 0 else f"{time_m:.0f}min"
        title += Text.format(" {B}", f" Tempo {time} ")
        title += Text.format(" {M}", f" Linhas {self.max_lines:.0f} ")
        title += Text.format(" {R}", f" Versões {self.versions} ")

        header: list[Text] = []
        # title = title.center(self.width)
        header.append(Text().add(first) + title.center(self.width - len(first)))
        
        if Flags.panel.get_value() == Flags.panel_logs:
            header.append(Text().addf("", second) + Text.format(" {C} {M}", " Scroll Up [PageUp] ", " Scroll Down [PgDown] ").center(self.width - len(second)))
            return header, self.raw_text
        
        if Flags.task_graph_mode.get_value() == Flags.task_time_view:
            exec_color = "Y"
            time_color = "G"
        else:
            exec_color = "G"
            time_color = "Y"
        header.append(Text().addf("", second) + Text.format(f" {{{exec_color}}} {{{time_color}}}", " ExecGraph [PageUp] ",  " TimeGraph [PageDown] ").center(self.width - len(second)))
        return header, self.get_graph()
        


import os
from pathlib import Path
from tko.settings.user_languages import UserLanguages

class Drafts:
    sandbox_key_prefix = "user"

    draft_readme: str = r"""Escreva aqui as informações que você quer salvar, esse é o seu rascunho.
O texto abaixo é informativo e você pode apagar depois de aprender como usar os rascunhos.

## Como usar os rascunhos

- Para renomear seu rascunho, basta renomear a pasta do rascunho.
- Para reorganizar, você pode criar subpastas dentro do sandbox e mover as pastas de rascunhos pra lá.
- Você pode usar o atalho R(Reload) no TKO para recarregar os rascunhos depois de criar, renomear ou reorganizar eles.
- Tudo que você fizer nos rascunhos também será rastreado pela ferramenta.

## Como criar seus próprios testes

Um teste é composto de um `input` (o texto que será fornecido para o programa) e um `output` (o texto que o programa deve retornar para esse input) e pode ter opcionalmente um `label` para facilitar a identificação do teste.

Seus casos de teste personalizados podem ser escritos diretamente aqui na descrição do problema dentro de um fenced code block com a linguagem `toml` ou em um arquivo `tests.toml` na pasta do problema. O TKO irá carregar automaticamente os testes quando a tarefa for aberta ou executada novamente.

Exemplo de teste para ler dois números, um por linha, e imprimir a soma e a subtração deles.

Se quiser habilitar esses casos de teste e ver funcionando, altere o fenced abaixo de `bash` para `toml` e execute novamente a tarefa no TKO.

```bash
[[tests]]
input = '''
3
2
'''
output = '''
5
1
'''

[[tests]]
label = 'números negativos'
input = '''
-3
-4
'''
output = '''
-7
1
'''
```

"""

    ts_draft = r"""
const input = () => ""; // MACRO
export {};
console.log("Hello, World!");
"""[1:]
    
    md_draft = r"""
# Rascunho

Se a tarefa exigir um relatório, escreva ele aqui. Você pode usar markdown, imagens e o que mais quiser para criar um relatório bem completo.
"""[1:]
    
    drafts = {
        'ts': ts_draft, 
        'md': md_draft
    }

    def get_languages_with_drafts(self) -> list[str]:
        keys = list(Drafts.drafts.keys())
        lang_register = UserLanguages(Path("")).set_default_path().load_file_settings().get_languages()
        for lang in lang_register.keys():
            if lang not in keys:
                keys.append(lang)
        keys.sort()
        return keys

    @staticmethod
    def create_draft_key(draft_id: int) -> str:
        return f"{Drafts.sandbox_key_prefix}_{draft_id:03d}"
    
    @staticmethod
    def find_max_numbered_key(task_keys_only: list[str]) -> int:
        numbered_keys: list[int] = []
        for key in task_keys_only:
            if key.startswith(Drafts.sandbox_key_prefix + "_"):
                try:
                    number = int(key[len(Drafts.sandbox_key_prefix) + 1:])
                    numbered_keys.append(number)
                except ValueError:
                    continue
        return max(numbered_keys) if numbered_keys else 0

    @staticmethod
    def load_drafts_only(folder: Path, lang: str, extra: list[str] | None = None) -> list[Path]:
        if extra is None:
            extra = []
        draft_list: list[Path] = []
        allowed = extra
        if lang != "":
            allowed.append(lang)
        if "c" in allowed:
            allowed.append("h")
        if "cpp" in allowed:
            allowed.append("h")
            allowed.append("hpp")
        if not os.path.isdir(folder):
            return []
        for root, _, files in os.walk(folder):
            cut_root = root[len(str(folder)):]
            pieces = cut_root.split(os.sep)
            if any([piece.startswith(".") for piece in pieces]) or any([piece.startswith("_") for piece in pieces]):
                continue
            for file in files:
                if file.endswith(tuple(allowed)):
                    draft_list.append(Path(os.path.join(root, file)))
        return draft_list
    
    @staticmethod
    def create_sandbox_draft(dir: Path, key: str):
        with open (os.path.join(dir, "README.md"), "w", encoding="utf-8") as f:
            f.write(f"---\nkey: {key}\n---\n\n# {os.path.basename(dir)}\n\n" + Drafts.draft_readme)



from __future__ import annotations
import argparse
import os
from tko.run.diff_builder_down import DiffBuilderDown
from tko.run.diff_builder_side import DiffBuilderSide
from tko.enums.diff_mode import DiffMode
from tko.util.raw_terminal import RawTerminal
from tko.run.unit import Unit


def cmd_diff(args: argparse.Namespace) -> None:
    target_a: str = args.target_a
    target_b: str = args.target_b
    diff_mode = DiffMode.SIDE if args.side else DiffMode.DOWN
    if args.path:
        if os.path.isfile(target_a):
            content_a = open(target_a, 'r', encoding='utf-8').read()
        else:
            content_a = "File not found: " + target_a
        if os.path.isfile(target_b):
            content_b = open(target_b, 'r', encoding='utf-8').read()
        else:
            content_b = "File not found: " + target_b
    else:
        content_a = target_a.replace('\\n', '\n')
        content_b = target_b.replace('\\n', '\n')
    unit: Unit = Unit()
    unit.set_expected(content_a)
    unit.set_received(content_b)
    if diff_mode == DiffMode.DOWN:
        diff_builder = DiffBuilderDown(RawTerminal.get_terminal_size(), unit).standalone_diff()
    else:
        diff_builder = DiffBuilderSide(RawTerminal.get_terminal_size(), unit).standalone_diff()
    for line in diff_builder.build_diff():
        print(line)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import argparse
import enum
from tko.feno.filter import Filter
from tko.util.decoder import Decoder


class Action(enum.Enum):
    RUN = 1
    CLEAN = 2

class TocMaker:
    @staticmethod
    def __only_hashtags(x: str) -> bool:
        return len(x) == x.count("#") and len(x) > 0

    # generate md link for the text
    @staticmethod
    def __get_md_link(title: str | None) -> str:
        if title is None:
            return ""
        # remove html comments
        if "<!--" in title and "-->" in title:
            title = title.split("<!--")[0]

        if "[](" in title:
            title = title.split("[](")[0]

        title = title.lstrip(" #")
        title = title.lower()
        out = ''
        for c in title:
            if c == ' ' or c == '-':
                out += '-'
            elif c == '_':
                out += '_'
            elif c == '\\':
                pass
            elif c.isalnum():
                out += c
        return out

    @staticmethod
    def __get_level(line: str) -> int:
        return len(line.split(" ")[0])

    @staticmethod
    def __get_content(line: str) -> str:
        if "<!--" in line and "-->" in line:
            line = line.split("<!--")[0]
        return " ".join(line.split(" ")[1:]).replace("\\", "\\\\")

    @staticmethod
    def __remove_code_fences(content: str) -> str:
        regex = r"^```.*?```\n"
        return re.sub(regex, "", content, 0, re.MULTILINE | re.DOTALL)


    @staticmethod
    def __extract_entries(content: str) -> list[tuple[int, str]]:
        content = TocMaker.__remove_code_fences(content)

        lines = content.splitlines()
        disable_tag = "[]()"
        lines = [line for line in lines if TocMaker.__only_hashtags(line.split(" ")[0]) and line.find(disable_tag) == -1]

        entries: list[tuple[int, str]] = []
        for line in lines:
            level = TocMaker.__get_level(line)
            text = "[" + TocMaker.__get_content(line) + "](#" + TocMaker.__get_md_link(line) + ")"
            entries.append((level, text))
        return entries

    
    @staticmethod
    def execute_toch(content: str) -> str:
        entries = TocMaker.__extract_entries(content)
        links = [b for (a, b) in entries if a == 2]
        table = ["--" for _ in links]
        return " | ".join(links) + "\n" + " | ".join(table)
        

    @staticmethod
    def execute_toc(content: str) -> str:
        entries = TocMaker.__extract_entries(content)
        toc_lines = ["  " * (level - 2) + "- " + link for (level, link) in entries if level > 1]
        toc_text = "\n".join(toc_lines)
        return toc_text

class Toc:
    @staticmethod
    def execute(content: str, action: Action = Action.RUN) -> str:
        regex = r"<!-- toc -->\n" + r"(.*?)"+ r"<!-- toc -->"
        if action == Action.RUN:
            new_toc = TocMaker.execute_toc(content)
            subst = r"<!-- toc -->\n" + new_toc + r"\n<!-- toc -->"
        else:
            subst = r"<!-- toc -->\n<!-- toc -->"
        return re.sub(regex, subst, content, 0, re.MULTILINE | re.DOTALL)

class Toch:
    @staticmethod
    def execute(content: str, action: Action = Action.RUN) -> str:
        regex = r"<!-- toch -->\n" + r"(.*?)"+ r"<!-- toch -->"
        if action == Action.RUN:
            new_toc = TocMaker.execute_toch(content)
            subst = r"<!-- toch -->\n" + new_toc + r"\n<!-- toch -->"
        else:
            subst = r"<!-- toch -->\n<!-- toch -->"
        return re.sub(regex, subst, content, 0, re.MULTILINE | re.DOTALL)

class Links:

    @staticmethod
    def load_links(readme_dir: str, filter_dir: str):
        def traverse_directory(directory: str, depth: int = 0):
            output = ""
            if os.path.isdir(directory):
                entries = sorted(os.listdir(directory))
                for entry in entries:
                    if entry.startswith("."):
                        continue
                    full_path = os.path.join(directory, entry)
                    if os.path.isdir(full_path):
                        output += "  " * depth + "- " + entry + "\n"
                        output += traverse_directory(full_path, depth + 1)
                    else:
                        path = os.path.relpath(full_path, readme_dir).replace("\\", "/")
                        output += "  " * depth + "- [" + entry + "](" + path + ")\n"
            return output
        
        origin = os.path.join(readme_dir, filter_dir)
        return traverse_directory(origin)

    @staticmethod
    def execute(path: str, content: str, action: Action = Action.RUN) -> str:
        regex = r"<!-- links (\S*?) -->\n(.*?)<!-- links -->"
        matches = re.finditer(regex, content, re.MULTILINE | re.DOTALL)
        
        # replace content of group 2 with load_links of group 1 for each match
        for match in matches:
            filter_dir = match.group(1)
            lregex = r"<!-- links " + filter_dir + r" -->\n(.*?)<!-- links -->"
            if action == Action.RUN:
                readme_dir = os.path.normpath(os.path.dirname(path))
                new_links = Links.load_links(readme_dir, filter_dir)
                subst = r"<!-- links " + filter_dir + r" -->\n" + new_links + r"<!-- links -->"
            else:
                subst = r"<!-- links " + filter_dir + r" -->\n<!-- links -->"
            content = re.sub(lregex, subst, content, 0, re.MULTILINE | re.DOTALL)

        return content

class Load:

    @staticmethod
    def extract_between_tags(content: str, tag: str) -> str:
        regex = r"\[\[" + tag + r"\]\].*?^(.*)^[\S ]*\[\[" + tag + r"\]\]"
        matches = re.finditer(regex, content, re.MULTILINE | re.DOTALL)
        for match in matches:
            return match.group(1)
        return ""

    @staticmethod
    def rmcom(target: str, content: str) -> str:
        com = "//"
        if target.endswith(".py"):
            com = "#"
        if target.endswith(".puml"):
            com = "'"
        lines = content.splitlines()
        output: list[str] = []
        for line in lines:
            if not line.lstrip().startswith(com):
                output.append(line)
        return "\n".join(output)


    @staticmethod
    def execute(content: str, target_dir: str, action: Action = Action.RUN) -> str:
        new_content = ""
        last = 0

        regex = r"<!-- load (\S*?) (\S*?) -->\n(.*?)<!-- load -->"
        matches = re.finditer(regex, content, re.MULTILINE | re.DOTALL)
        
        for match in matches:
            path = match.group(1)
            abspath = os.path.abspath(os.path.join(target_dir, path))
            tags = match.group(2)
            words: list[str] = tags.split(":")

            fenced: list[str] = [tag for tag in words if tag.startswith("fenced")]
            words = [tag for tag in words if not tag.startswith("fenced")]

            filter: list[str] = [tag for tag in words if tag.startswith("filter")]
            words = [tag for tag in words if not tag.startswith("filter")]

            rmcom: list[str] = [tag for tag in words if tag.startswith("rmcom")]
            words = [tag for tag in words if not tag.startswith("rmcom")]


            extract: list[str] = [tag for tag in words if tag.startswith("extract")]
            words = [tag for tag in words if not tag.startswith("extract")]


            ext = os.path.splitext(path)[1][1:]
            if len(words) > 0:
                ext = words[0]
            if len(fenced) == 1:
                parts = fenced[0].split("=")
                if len(parts) == 2:
                    ext = parts[1]


            new_content += content[last:match.start()] # inserindo texto entre matches
            last = match.end()
            # new_content += "[](load)[](" + path + ")[](" + tags + ")\n"
            new_content += "<!-- load " + path + " " + tags + " -->\n"

            # se não for run, deve limpar o conteúdo não inserindo os arquivos
            if action == Action.RUN:
                if len(fenced) > 0:
                    new_content += "\n```" + ext + "\n"
                if os.path.isfile(abspath):
                    if len(filter) > 0:
                        data = Filter(path).process(Decoder.load(abspath)) + "\n"
                        new_content += data
                        if data[-1] != "\n":
                            new_content += "\n"
                    elif len(rmcom) > 0:
                        full_data = Decoder.load(abspath)
                        data =  Load.rmcom(abspath, full_data) + "\n"
                        new_content += data
                        if data[-1] != "\n":
                            new_content += "\n"
                    elif len(extract) > 0:
                        tag = extract[0].split("=")[1]
                        full_data = Decoder.load(abspath)
                        data = Load.extract_between_tags(full_data, tag)
                        new_content += data
                        if len(data) == 0 or data[-1] != "\n":
                            new_content += "\n"
                    else:
                        data = Decoder.load(abspath)
                        new_content += data
                        if data[-1] != "\n":
                            new_content += "\n"
                else:
                    print("warning: file", path, "not found")
                if fenced:
                    new_content += "```\n\n"
            new_content += "<!-- load -->"
        
        new_content += content[last:]
        return new_content


            

class Save:
    @staticmethod
    # execute filename and content
    def execute(file_content: str) -> None:
        regex = r"\[\]\(save\)\[\]\((.*?)\)\n```[a-z]*\n(.*?)```\n\[\]\(save\)"
        matches = re.finditer(regex, file_content, re.MULTILINE | re.DOTALL)
        content_old = ""        
        for match in matches:
            path = match.group(1)
            content = match.group(2)
            exists = os.path.isfile(path)
            if exists:
                content_old = Decoder.load(path)
            if not exists or content != content_old:
                Decoder.save(path, content)
                print("file", path, "updated")

class MdppMain:
    @staticmethod
    def fix_path(target: str):
        target = os.path.normpath(target)
        if os.path.isdir(target):
            target = os.path.join(target, "README.md")
        return target

    @staticmethod
    def open_file(path: str) -> tuple[bool, str]: 
        if os.path.isfile(path):
            file_content = Decoder.load(path)
            return True, file_content
        print("Warning: File", path, "not found")
        return False, "" 

class Mdpp:
    @staticmethod
    def update_file(target: str, action: Action = Action.RUN, quiet: bool = False) -> bool:
        # path = MdppMain.fix_path(target)
        path = target
        if not path.endswith(".md"):
            print("Warning: File", path, "is not a markdown file")
            return False
        if not os.path.isfile(path):
            print("Warning: File", path, "not found")
            return False
        target_dir = os.path.dirname(path)
        found, original = MdppMain.open_file(path)
        if not found:
            return False
        updated = original
        updated_toc = Toc.execute(updated, action)
        updated_toc = Toch.execute(updated_toc, action)
        if updated != updated_toc:
            updated = updated_toc
        updated = Load.execute(updated, target_dir, action)
        updated = Links.execute(target, updated, action)
        Save.execute(updated)
        
        if updated != original:
            Decoder.save(path, updated)
            # hook = os.path.abspath(path).split(os.sep)[-2]
            return True

        return False

def mdpp_main(args: argparse.Namespace):

    targets: list[str] = args.targets
    if len(args.targets) == 0:
        targets = [os.path.join(os.getcwd(), "README.md")]
        print(f"Updating README.md in {os.path.basename(os.getcwd())}")

    action = Action.RUN if not args.clean else Action.CLEAN

    for target in targets:
        Mdpp.update_file(target, action, args.quiet)



from typing import Tuple
import os
import glob
from os.path import getmtime
import argparse


class Older:
    # return the last update for the most recent file in the directory
    @staticmethod
    def last_update(path: str) -> Tuple[str, float]:
        value = None
        if os.path.isfile(path):
            value = (path, getmtime(path))
        else:
            file_list = list(glob.iglob(path + '/**', recursive=True))
            file_list = [f for f in file_list if os.path.isfile(f)]
            if len(file_list) == 0:
                value = (path, getmtime(path))
            else:
                juntos = [(f, getmtime(f)) for f in file_list]
                value = max(juntos, key=lambda x: x[1])
        return value

    # busca recursivamente o arquivo mais recente dentro os arquivos passados
    @staticmethod
    def find_older(targets: list[str]) -> str:
        best_file, best_time = Older.last_update(targets[0])
        for target in targets[1:]:
            file, time = Older.last_update(target)
            if time > best_time:
                best_file, best_time = file, time
        return best_file

def older_main(args: argparse.Namespace):
    print(Older.find_older(args.targets))



from __future__ import annotations
from pathlib import Path
from tko.settings.rep_data import RepData
from typing import Any
from tko.settings.rep_data import RepData
from tko.settings.rep_source import RepSource
from tko.game.game import Game
from tko.util.decoder import Decoder
from tko.util.text import Text
from tko.logger.logger import Logger
import yaml # type: ignore
from tko.settings.rep_paths import RepPaths
from icecream import ic # type: ignore
from datetime import timedelta
from tko.settings.git_cache import GitCache
import os

def remove_git_merge_tags(lines: list[str]) -> list[str]:
    # remove lines with <<<<<<<, =======, >>>>>>>>
    filtered_lines: list[str] = []
    for line in lines:
        if line.startswith("<<<<<<<") or line.startswith("=======") or line.startswith(">>>>>>>"):
            continue
        filtered_lines.append(line)
    return filtered_lines

class Repository:
    cache_time_for_remote_source = 3600 # seconds

    def __init__(self, folder: Path, force_update: bool = False):
        rep_folder: Path = folder
        recursive_folder = RepPaths.rec_search_for_repo_parents(folder)
        if recursive_folder is not None:
            rep_folder = recursive_folder
        self.paths = RepPaths(rep_folder)
        self.data: RepData = RepData()
        self.game = Game()
        self.logger: Logger = Logger(rep_folder) 
        self.force_update: bool = force_update
        self.cache = GitCache(self.paths.get_cache_folder(), timedelta(seconds=self.cache_time_for_remote_source))

    def set_global_cache(self):
        RepPaths.use_global_cache_folder = True
        self.cache = GitCache(self.paths.get_cache_folder(), timedelta(seconds=self.cache_time_for_remote_source))
        return self


    def found(self):
        return self.paths.get_config_file().exists()

    def is_local_dir(self, path: Path) -> bool:
        rep_dir = self.paths.get_workspace_dir()
        path = path.resolve()
        return path.is_relative_to(rep_dir)

    def load_game(self, try_update: bool = True, silent: bool = False) -> Repository:
        if not self.data.get_sources():
            self.load_config()
        if try_update:
            for source in self.data.get_sources():
                _ = source.get_source_readme() # to ensure cache path is set
        sources: list[RepSource] = self.data.get_sources()
        self.game.set_sources(sources, self.data.lang, silent=silent)
        self.__load_tasks_from_rep_into_game()
        return self
    
    def get_key_from_task_folder(self, folder: Path) -> str:
        path = folder.resolve()
        parts = path.parts
        label = parts[-1]
        workspace = parts[-2]
        return f"{workspace}@{label}"

    def is_task_folder(self, folder: Path) -> bool:
        label = folder.name
        task_folder = self.get_task_folder_for_label(label)
        return task_folder == folder
    
    def __load_tasks_from_rep_into_game(self):
        # load tasks from repository.yaml into game
        tasks = self.data.tasks
        for key, serial in tasks.items():
            if key in self.game.tasks:
                self.game.tasks[key].load_from_db(serial)

    def get_task_folder_for_label(self, label: str) -> Path:
        parts: list[str] = label.split("@")
        source = ""
        if len(parts) > 1:
            source = parts[0]
            label = parts[1]
        return self.paths.root_folder / source / label


    def load_config(self):
        content = Decoder.load(self.paths.get_config_file())
        local_data: dict[str, Any] = {}
        try:
            lines = content.splitlines()
            lines = remove_git_merge_tags(lines)
            local_data = yaml.safe_load("\n".join(lines))
            if local_data is None or not isinstance(local_data, dict) or len(local_data) == 0: # type: ignore
                backup_content = Decoder.load(self.paths.get_config_backup_file())
                lines = backup_content.splitlines()
                lines = remove_git_merge_tags(lines)
                local_data = yaml.safe_load("\n".join(lines))
            if local_data is None or not isinstance(local_data, dict) or len(local_data) == 0: # type: ignore
                raise FileNotFoundError(f"Arquivo de configuração vazio: {self.paths.get_config_file()}")
        except:
            raise Warning(Text.format("O arquivo de configuração do repositório {y} está {r}.\nAbra e corrija o conteúdo ou crie um novo.", self.paths.get_config_file(), "corrompido"))
        self.data.load_from_dict(local_data)
        self.data.ensure_sandbox_source(self.paths.get_workspace_dir())

        for source in self.data.get_sources():
            source.set_rep_globals(self.paths.get_workspace_dir(), self.paths.get_cache_folder())

        return self

    def get_student_sandbox(self) -> RepSource:
        source = RepSource("").set_student_sandbox()
        source.set_rep_globals(self.paths.get_workspace_dir(), self.paths.get_cache_folder())
        return source


    def save_config(self):
        self.data.version = "0.2"

        path: Path = Path(self.paths.get_config_file())
        path.parent.mkdir(parents=True, exist_ok=True)
        atomic_write_yaml(path, self.data.save_to_dict())
        return self

    def __str__(self) -> str:
        return f"data: {self.data}\n"
        

def atomic_write_yaml(path: Path, data: dict[str, Any]) -> None:
    tmp: Path = path.with_suffix(path.suffix + ".tmp")

    with tmp.open("w", encoding="utf-8") as f:
        yaml.safe_dump(data, f, allow_unicode=True, sort_keys=False)

        f.flush()
        os.fsync(f.fileno())

    tmp.replace(path)

    # fsync do diretório (POSIX apenas)
    if os.name == "posix":
        dir_fd: int = os.open(path.parent, os.O_DIRECTORY)
        try:
            os.fsync(dir_fd)
        finally:
            os.close(dir_fd)


from tko.game.task import Task
from tko.game.quest import Quest
from tko.play.tasktree import TaskTree
from tko.play.flags import Flags
from tko.play.floating_manager import FloatingManager
import curses

class Search:
    def __init__(self, tree: TaskTree, fman: FloatingManager):
        self.tree = tree
        self.game = tree.game
        self.fman = fman
        self.search_mode: bool = False
        self.backup_expanded: list[str] = []
        self.backup_index_selected: str = ""
        self.settings = tree.settings

    def toggle_search(self):
        if not self.search_mode:
            self.backup_expanded = [v for v in self.tree.expanded]
            self.backup_index_selected = self.tree.selected_item
            Flags.tasks.set_value("0")
            self.tree.update_tree(admin_mode=True)
            self.tree.process_expand_all()
            self.tree.process_expand_all()
        self.search_mode = not self.search_mode
            
    def finish_search(self):
        if self.tree.selected_item == "":
            self.cancel_search()
            return

        self.search_mode = False
        self.tree.search_text = ""
        selected_key = self.tree.selected_item
        item = self.tree.all_items[selected_key]
        reachable = True
        if isinstance(item, Task):
            reachable = self.game.quests[item.quest_key].is_reachable()
        elif isinstance(item, Quest):
            reachable = item.is_reachable()
        if not reachable:
            Flags.inbox.set_value(Flags.inbox_all)
        self.tree.update_tree(Flags.inbox.get_value() == Flags.inbox_all) # usa o mode de antes e vê se acha
        self.tree.reload_sentences()
        
        self.tree.expanded = set()
        unit = self.tree.all_items[selected_key]

        if isinstance(unit, Task):
            self.tree.expanded = {unit.remote_name, unit.quest_key}
        elif isinstance(unit, Quest):
            self.tree.expanded = {unit.get_full_key(), unit.remote_name}
        self.tree.reload_sentences()

    # update index to match the first item that matches the search
    def update_index(self):
        _, first = self.tree.filter_by_search()
        self.tree.selected_item = first if first is not None else ""

    def cancel_search(self):
            self.search_mode = False
            self.tree.search_text = ""
            self.tree.expanded = {v for v in self.backup_expanded}
            self.tree.selected_item = self.backup_index_selected

    def process_search(self, key: int):
        if key == curses.KEY_EXIT:
            self.cancel_search()
        elif key == ord("\n"):
            self.finish_search()
        elif key == curses.KEY_UP:
            self.tree.move_up()
        elif key == curses.KEY_DOWN:
            self.tree.move_down()
        elif key == curses.KEY_LEFT or key == curses.KEY_BACKSPACE:
            self.tree.search_text = self.tree.search_text[:-1]
            self.update_index()
        elif 32 <= key < 127:
            self.tree.search_text += chr(key).lower()
            self.update_index()



import math
import os

from tko.util.identifier import Identifier
from tko.enums.identifier_type import IdentifierType
from tko.run.unit import Unit
from tko.util.param import Param
from tko.run.loader import Loader
from tko.run.solver_builder import SolverBuilder
from tko.util.text import Text
from tko.util.symbols import symbols
from tko.util.label_factory import LabelFactory
from tko.down.drafts import Drafts
from pathlib import Path
from tko.settings.settings import Settings

class Wdir:
    def __init__(self, settings: Settings):
        self.settings = settings
        self.__autoload = False
        self.__autoload_folder: Path | None = None
        self.__solver: None | SolverBuilder = None
        self.__source_list: list[Path] = []
        self.__pack_list: list[list[Unit]] = []
        self.__unit_list: list[Unit] = []
        self.__curses = False
        self.__lang = ""

    def has_solver(self) -> bool:
        return not self.__solver is None

    def has_tests(self) -> bool:
        return len(self.__unit_list) != 0

    def get_solver(self) -> SolverBuilder:
        if self.__solver is None:
            raise Warning("fail: Não foi encontrado arquivo de código")
        return self.__solver
    
    def get_unit_list(self) -> list[Unit]:
        return self.__unit_list

    def get_unit(self, index: int) -> Unit:
        return self.__unit_list[index]
    
    def get_source_list(self) -> list[Path]:
        return self.__source_list

    def set_curses(self, value: bool):
        self.__curses = value
        return self

    def set_lang(self, lang: str):
        self.__lang = lang
        return self
    
    def is_curses(self) -> bool:
        return self.__curses

    def is_autoload(self) -> bool:
        return self.__autoload

    def get_autoload_folder(self) -> Path | None:
        return self.__autoload_folder

    def set_solver(self, solver_list: list[Path]):
        if len(solver_list) > 0:
            self.__solver = SolverBuilder(solver_list, self.settings)
        return self

    def set_sources(self, source_list: list[Path]):
        self.__source_list = source_list
        return self

    def autoload(self):
        if self.__autoload_folder is None:
            raise Warning("fail: pasta de autoload não definida")
        folder: Path = self.__autoload_folder

        # loading source list
        source_list: list[Path] = [target for target in folder.iterdir() if target.suffix in [".tio", ".vpl", ".toml"]]
        source_list.extend([target for target in folder.iterdir() if target.suffix == ".md"])
        
        if self.__lang != "":
            solver_list = Drafts.load_drafts_only(folder, self.__lang)
        else:
            solver_list = Drafts.load_drafts_only(folder, "", extra=Drafts().get_languages_with_drafts())
        solver_list = sorted(solver_list)

        self.set_solver(solver_list)
        self.set_sources(source_list)
        self.__autoload = True
        return self

    def set_target_list(self, target_list: list[Path]):
        if len(target_list) == 0:
            target_list.append(Path())
        if len(target_list) == 1 and os.path.isdir(target_list[0]):
            self.__autoload_folder = target_list[0]
            return self.autoload()
                    
        for target in target_list:
            if not os.path.exists(target):
                raise Warning(f"fail: {target} não encontrado")

        solvers = [target for target in target_list if Identifier.get_type(target.suffix) == IdentifierType.SOLVER]
        sources = [target for target in target_list if Identifier.get_type(target.suffix) != IdentifierType.SOLVER]

        self.set_solver(solvers)
        self.set_sources(sources)
        return self

    def build(self):
        loading_failures = 0
        self.__pack_list = []
        for source in self.__source_list:
            try:
                raw_list: list[Unit] = Loader.parse_source(source)
                unit_list: list[Unit] = []
                for unit in raw_list:
                    if unit.get_expected() == "" and unit.get_input() == "":
                        continue
                    unit_list.append(unit)
                self.__pack_list.append(unit_list)
            except FileNotFoundError as e:
                print(str(e))
                loading_failures += 1
                pass
        if loading_failures > 0 and loading_failures == len(self.__source_list):
            raise FileNotFoundError("failure: nenhum arquivo de teste encontrado")
        self.__unit_list = []
        input_dict: dict[str, int] = {}
        index = 0
        for pack in self.__pack_list:
            for unit in pack:
                if unit.get_input() in input_dict:
                    unit.repeated = input_dict[unit.get_input()]
                else:
                    input_dict[unit.get_input()] = index
                    unit.index = index
                    self.__unit_list.append(unit)
                    index += 1
        #self.__number_and_mark_duplicated()
        #self.__remove_duplicated()
        self.__calculate_grade()
        self.__pad()
        return self

    def calc_grade(self) -> int:
        grade = 100
        for case in self.__unit_list:
            if not case.repeated and (case.get_received() is None or case.get_expected() != case.get_received()):
                grade -= case.grade_reduction
        return max(0, grade)

    # put all the labels with the same length
    def __pad(self):
        if len(self.__unit_list) == 0:
            return
        max_case = max([len(x.case) for x in self.__unit_list])
        max_source = max([len(x.source) for x in self.__unit_list])
        for unit in self.__unit_list:
            unit.case_pad = max_case
            unit.source_pad = max_source

    # select a single unit to execute exclusively
    def filter(self, param: Param.Basic):
        index = param.index
        if index is not None:
            if 0 <= index < len(self.__unit_list):
                self.__unit_list = [self.__unit_list[index]]
            else:
                raise ValueError("Índice fora dos limites: " + str(index))
        return self

    # calculate the grade reduction for the cases without grade
    # the grade is proportional to the number of unique cases
    def __calculate_grade(self):
        unique_count = len([x for x in self.__unit_list if not x.repeated])
        for unit in self.__unit_list:
            if unit.grade is None:
                unit.grade_reduction = math.floor(100 / unique_count)
            else:
                unit.grade_reduction = unit.grade

    # number the cases and mark the repeated
    # def __number_and_mark_duplicated(self):
    #     new_list: list[Unit] = []
    #     index = 0
    #     for unit in self.__unit_list:
    #         unit.index = index
    #         index += 1
    #         search = [x for x in new_list if x.input == unit.input]
    #         if len(search) > 0:
    #             unit.repeated = search[0].index
    #         new_list.append(unit)
    #     self.__unit_list = new_list

    # def __remove_duplicated(self):
    #     self.__unit_list = [unit for unit in self.__unit_list if unit.repeated is None]

    # sort, unlabel ou rename using the param received
    def manipulate(self, param: Param.Manip):
        # filtering marked repeated
        self.__unit_list = [unit for unit in self.__unit_list if unit.repeated is None]
        if param.to_sort:
            self.__unit_list.sort(key=lambda v: len(v.input))
        if param.unlabel:
            for unit in self.__unit_list:
                unit.case = ""
        if param.to_number:
            number = 00
            for unit in self.__unit_list:
                unit.case = LabelFactory().label(unit.case).index(number).generate()
                number += 1

    def unit_list_resume(self) -> list[Text]:
        return [unit.str() for unit in self.__unit_list]

    def sources_names(self) -> list[tuple[str, str]]:
        out: list[tuple[str, str]] = []
        if len(self.__pack_list) == 0:
            out.append((symbols.failure.text, "0"))
        for source_name, ulist in zip(self.__source_list, self.__pack_list):
            nome: str = source_name.name
            count = len([unit for unit in ulist if unit.repeated is None])
            if count > 0:
                count_str = str(count)
                total = len(ulist)
                if count < total:
                    count_str += "/" + str(total)
                out.append((nome, count_str))
        return out

    def solvers_names(self) -> list[str]:
        path_list = [] if self.__solver is None else self.__solver.args_list
        if self.__solver is not None and len(path_list) == 0:  # free_cmd
            out = ["free cmd"]
        else:
            out = [os.path.basename(path) for path in path_list]
        return out

    def resume_splitted(self) -> Text:
        sources = ["{}({})".format(name, str(count).rjust(2, "0")) for name, count in self.sources_names()]
        __sources = Text().add("Testes:").add("[").addf("y", ", ".join(sources)).add("]")
        __solvers = Text().add("Códigos:").add("[").addf("g", ", ".join(self.solvers_names())).add("]")
        return Text().add(__solvers).add(" ").add(__sources)

    def resume_join(self) -> Text:
        sources = ["{}({})".format(name, str(count).rjust(2, "0")) for name, count in self.sources_names()]
        __sources = Text().addf("y", ", ".join(sources))
        __solvers = Text().addf("g", ", ".join(self.solvers_names()))
        return Text().add(__solvers).add(", ").add(__sources)


from tko.util.text import Text

class __Symbols:

    def __init__(self):
        self.task_downloaded = Text.Token("▼", "g")
        self.task_to_download = Text.Token("▲")
        self.task_local = Text.Token("▶", "g")
        
        self.opening = Text.Token("▶")
        self.right_arrow = Text.Token("▶")
        self.left_arrow = Text.Token("◀")
        self.neutral = Text.Token("»")
        self.success = Text.Token("✓")
        self.failure = Text.Token("✗")
        self.wrong = Text.Token("ω")
        self.compilation = Text.Token("ϲ")
        self.execution = Text.Token("ϵ")
        self.unequal = Text.Token("├")
        self.equalbar = Text.Token("│")
        self.hbar = Text.Token("─")
        self.vbar = Text.Token("│")

        self.execution_result: dict[str, Text.Token] = {
            "untested": self.neutral,
            "success": self.success,
            "wrong_output": self.wrong,
            "compilation_error": self.compilation,
            "execution_error": self.execution,
        }

        self.whitespace = Text.Token("·") #Text.Token("␣")
        self.newline = Text.Token("↲")

        self.circle_filled = Text.Token("●")
        self.circle_open = Text.Token("○")

        self.square_filled = Text.Token("■")
        self.square_open = Text.Token("□")

        self.task_human_filled = Text.Token("◆")
        self.task_human_open = Text.Token("◇")

        self.star_filled = Text.Token("★")
        self.star_open = Text.Token("☆")

        self.task_denied = Text.Token("⊘")
        self.task_repeat = Text.Token("↺")

        self.task_edit = Text.Token("✎")
        self.task_view = Text.Token("↗")


        self.cfill = Text.Token("_")
        self.tab = Text.Token("    ")
        self.arrow_up = Text.Token("↑")

        self.check = Text.Token("✓")
        self.uncheck = Text.Token("✗")

        self.infinity = Text.Token("∞")
        self.locked_free = Text.Token("⇉")
        self.locked_locked = Text.Token("⇟")
        self.left_toggle = Text.Token("━─")
        self.right_toggle = Text.Token("─━")
        self.timer = Text.Token("⏳") #⏰
        self.diff_down = Text.Token("↓")
        self.diff_left = Text.Token("→")
        self.sharp_r = Text.Token("")
        self.sharpL = Text.Token("")
        self.action = Text.Token("◎", "b")

        # self.emoji_confiante = Text.Token("●", "g")
        # self.emoji_capaz     = Text.Token("◕", "y")
        # self.emoji_inseguro  = Text.Token("◑", "m")
        # self.emoji_confuso   = Text.Token("◔", "r")
        # self.emoji_nao_fiz   = Text.Token("ⵔ")

        # self.emoji_confiante = Text.Token("■", "g")
        # self.emoji_capaz     = Text.Token("◨", "y")
        # self.emoji_inseguro  = Text.Token("◧", "m")
        # self.emoji_confuso   = Text.Token("◱", "r")
        # self.emoji_nao_fiz   = Text.Token("□")

        self.edge_a = Text.Token("▇", "g")
        self.edge_b = Text.Token("▆", "y")
        self.edge_c = Text.Token("▅", "m")
        self.edge_d = Text.Token("▄", "r")
        self.edge_e = Text.Token("▃", "c")
        self.edge_x = Text.Token("▂", "")
        self.edge_list = [self.edge_x, self.edge_e, self.edge_d, self.edge_c, self.edge_b, self.edge_a]

        self.flow_s = Text.Token("S", "b")
        self.flow_a = Text.Token("A", "g")
        self.flow_b = Text.Token("P", "y")
        self.flow_c = Text.Token("G", "y")
        self.flow_d = Text.Token("I", "y")
        self.flow_e = Text.Token("H", "y")
        self.flow_x = Text.Token("x", "")
        self.flow_list = [self.flow_x, self.flow_e, self.flow_d, self.flow_c, self.flow_b, self.flow_a, self.flow_s]
        
        self.cool_a = Text.Token("5", "g")
        self.cool_b = Text.Token("4", "y")
        self.cool_c = Text.Token("3", "m")
        self.cool_d = Text.Token("2", "r")
        self.cool_e = Text.Token("1", "c")
        self.cool_x = Text.Token("x", "")
        self.cool_list = [self.cool_x, self.cool_e, self.cool_d, self.cool_c, self.cool_b, self.cool_a]

        self.star = Text.Token("★", "g")
        self.open_star = Text.Token("☆")
        self.mark = Text.Token("✦", "g")

        self.cursor = Text.Token("┊")

    # def set_ascii(self):
    #     self.task_downloaded = Text.Token("D", "g")
    #     self.task_to_download = Text.Token("X")
    #     self.task_local = Text.Token(">", "g")
    #     self.whitespace = Text.Token("¨")
    #     self.newline = Text.Token("~")

    #     self.cfill = Text.Token("_")
    #     self.tab = Text.Token("    ")
    #     self.arrow_up = Text.Token("|")

    #     self.check = Text.Token("A")
    #     self.uncheck = Text.Token("0")

    #     self.infinity = Text.Token("0")
    #     self.locked_free = Text.Token(">")
    #     self.locked_locked = Text.Token("v")
    #     self.left_toggle = Text.Token("━─")
    #     self.right_toggle = Text.Token("─━")
    #     self.timer = Text.Token("l")
    #     self.diff_down = Text.Token("|")
    #     self.diff_left = Text.Token("─")
    #     self.sharp_r = Text.Token("")
    #     self.sharpL = Text.Token("")
    #     self.action = Text.Token("◎", "b")

    #     self.emoji_vazio = Text.Token("✗", "r")
    #     self.emoji_confiante = Text.Token("A", "g")
    #     self.emoji_capaz     = Text.Token("B", "y")
    #     self.emoji_inseguro  = Text.Token("C", "m")
    #     self.emoji_confuso   = Text.Token("D", "r")
    #     self.emoji_nao_fiz   = Text.Token("E")

    #     self.emoji_alone = Text.Token("A", "g")
    #     self.emoji_dicas = Text.Token("B", "y")
    #     self.emoji_codes = Text.Token("C", "m")
    #     self.emoji_guide = Text.Token("D", "r")
    #     self.cursor = Text.Token("|")

    def set_colors(self):
        self.opening.fmt = "b"
        self.neutral.fmt = "b"
        self.success.fmt = "g"
        self.failure.fmt = "r"
        self.wrong.fmt = "r"
        self.compilation.fmt = "y"
        self.execution.fmt = "y"
        self.unequal.fmt = "r"
        self.equalbar.fmt = "g"

symbols = __Symbols()



from __future__ import annotations
from tko.logger.log_item_base import LogItemBase

import enum

class LogItemExec(LogItemBase):
    class Fail(enum.Enum):
        NONE = "NONE"
        COMP = "COMP"
        EXEC = "EXEC"

    class Mode(enum.Enum):
        NONE = "NONE" # no execution
        FULL = "FULL" # run all test cases
        LOCK = "LOCK" # run just one test case  
        FREE = "FREE" # run without test cases

    mode_str = "mode"
    rate_str = "rate"
    size_str = "size"
    fail_str = "fail"
    def __init__(self):
        super().__init__(LogItemBase.Type.EXEC)
        self.mode: LogItemExec.Mode = LogItemExec.Mode.NONE # NONE, FULL, LOCK, FREE
        self.rate: int = -1 # percentage of coverage, value from 0 to 100
        self.size: int = -1 # lines
        self.fail: LogItemExec.Fail = LogItemExec.Fail.NONE

    def set_mode(self, mode: LogItemExec.Mode):
        self.mode = mode
        return self

    def set_rate(self, cov: int):
        self.rate = cov
        return self

    def set_size(self, changes: bool, lines: int):
        if changes:
            self.size = lines
        return self

    def set_fail(self, err: LogItemExec.Fail):
        self.fail = err
        return self

    def get_rate(self) -> int:
        return self.rate

    def get_size(self) -> int:
        return self.size

    def get_fail(self) -> LogItemExec.Fail:
        return self.fail

    def identify_kv(self, kv: dict[str, str]) -> bool:
        if self.rate_str in kv:
            self.rate = int(kv[self.rate_str])
        if self.size_str in kv:
            self.size = int(kv[self.size_str])
        if self.mode_str in kv:
            self.mode = LogItemExec.Mode(kv[self.mode_str])
        if self.fail_str in kv:
            self.fail = LogItemExec.Fail(kv[self.fail_str])
        
        if self.key == "":
            return False
        return True

    def encode_line(self) -> str:
        output = super().encode_line()
        if self.mode != LogItemExec.Mode.NONE:
            output += f', {self.mode_str}:{self.mode.value}'
        if self.rate >= 0:
            output += f', {self.rate_str}:{self.rate}'
        if self.size >= 0:
            output += f', {self.size_str}:{self.size}'
        if self.fail != LogItemExec.Fail.NONE:
            output += f', {self.fail_str}:{self.fail.value}'
        return output
    



from __future__ import annotations
import re


class FileSource:
    def __init__(self, label: str, input_file: str, output_file: str):
        self.label = label
        self.input_file = input_file
        self.output_file = output_file

    def __eq__(self, other: FileSource) -> bool: # type: ignore
        return self.label == other.label and self.input_file == other.input_file and \
                self.output_file == other.output_file


class PatternLoader:
    pattern: str = ""

    def __init__(self):
        parts = PatternLoader.pattern.split(" ")
        self.input_pattern = parts[0]
        self.output_pattern = parts[1] if len(parts) > 1 else ""
        self._check_pattern()

    def _check_pattern(self):
        self.__check_double_wildcard()
        self.__check_missing_wildcard()

    def __check_double_wildcard(self):
        if self.input_pattern.count("@") > 1 or self.output_pattern.count("@") > 1:
            raise ValueError("  fail: the wildcard @ should be used only once per pattern")

    def __check_missing_wildcard(self):
        if "@" in self.input_pattern and "@" not in self.output_pattern:
            raise ValueError("  fail: is input_pattern has the wildcard @, the input_patter should have too")
        if "@" not in self.input_pattern and "@" in self.output_pattern:
            raise ValueError("  fail: is output_pattern has the wildcard @, the input_pattern should have too")

    def make_file_source(self, label: str) -> FileSource:
        return FileSource(label, self.input_pattern.replace("@", label), self.output_pattern.replace("@", label))

    def get_file_sources(self, filename_list: list[str]) -> list[FileSource]:
        input_re = self.input_pattern.replace(".", "\\.")
        input_re = input_re.replace("@", "(.*)")
        file_source_list: list[FileSource] = []
        for filename in filename_list:
            match = re.findall(input_re, filename)
            if not match:
                continue
            label = match[0]
            file_source = self.make_file_source(label)
            if file_source.output_file not in filename_list:
                print("fail: file " + file_source.output_file + " not found")
            else:
                file_source_list.append(file_source)
        return file_source_list

    def get_odd_files(self, filename_list: list[str]) -> list[str]:
        matched: list[str] = []
        sources: list[FileSource] = self.get_file_sources(filename_list)
        for source in sources:
            matched.append(source.input_file)
            matched.append(source.output_file)
        unmatched = [file for file in filename_list if file not in matched]
        return unmatched



class QuestGrader:
    class Elem:
        def __init__(self, opt: bool, value: int, percent: float):
            self.opt = opt
            self.value = value
            self.percent = percent

    @staticmethod
    def calc_xp_earned_total(tasks: list[Elem]) -> tuple[float, float]:
        """
        tasks: list of tuples (opt, xp, percent)
        """
        total_xp = 0.0
        earned_xp = 0.0
        for elem in tasks:
            total_xp += elem.value
            if elem.percent > 1:
                earned_xp += elem.value * (elem.percent / 100.0)
        return earned_xp, total_xp
    
    @staticmethod
    def get_percent(earned_xp: float, total_xp: float) -> float:
        if total_xp == 0:
            return 0.0
        return (earned_xp * 100.0) / total_xp


from tko.settings.settings import Settings
from tko.settings.repository import Repository
from tko.play.play import Play
from tko.util.text import Text
from pathlib import Path

class CmdOpen:
    def __init__(self, settings: Settings, force_update: bool):
        self.settings = settings
        self.need_update = False
        self.repo: Repository | None = None
        self.repo_dir: Path | None = None
        self.force_update = force_update

    def display_need_update(self):
        self.need_update = True

    def load_folder(self, repo_dir: Path):
        self.repo_dir = repo_dir
        self.repo = Repository(repo_dir, self.force_update)
        if not self.repo.paths.has_local_config_file():
            print(Text.format("{r}: O comando {g} deve ser executado na pasta onde o repositório foi iniciado.", "Erro", "tko open"))
            print(Text.format("{g}: Navegue ou passe o caminho até a pasta do repositório e tente novamente.", "Ação"))
            raise Warning(Text.format("{r}: {y} {}", "Erro:", repo_dir, "não contém um repositório do tko"))
        self.repo.load_config().load_game()
        return self

    def execute(self):
        if self.repo is None:
            raise Warning("Repositório não encontrado")

        play = Play(self.settings, self.repo)
        if self.need_update:
            play.display_need_update()
        play.play()



import datetime
import os
import argparse
import csv
from pathlib import Path
from tko.play.patch_history import PatchHistory
from tko.util.decoder import Decoder
import tempfile
from tko.logger.log_sort import LogSort
from pathlib import Path

class Track:
    def __init__(self):
        self.timestamp = ""
        self.file_stamp_list: list[str] = []
        self.result: str = ""

    def set_timestamp(self, timestamp: str):
        self.timestamp = timestamp
        return self
    
    def set_file_stamp_list(self, files: list[str]):
        self.file_stamp_list = [os.path.basename(f) for f in files]
        return self
    
    def set_result(self, result: str):
        self.result = result
        return self
    
    def track_to_column(self):
        return [self.timestamp, self.result, ";".join(self.file_stamp_list)]

    def column_to_track(self, columns: list[str]):
        if len(columns) < 3:
            raise ValueError("Not enough columns to create a Track object.")
        self.timestamp = columns[0]
        self.result = columns[1]
        self.file_stamp_list = columns[2].split(";")
        return self


class Tracker:
    log_file = "track.csv"
    extension = ".json"

    def __init__(self):
        self._result: str = "None"
        self._files: list[Path] = []
        self._folder: Path = Path()

    def unfold_files(self, log_sort: LogSort) -> tuple[str, str]:
        output = "\n"


        timestamp_rate: dict[str, str] = {}
        for _, item in log_sort.exec_list:
            timestamp = item.get_timestamp().replace(":", "-").replace(" ", "_")
            timestamp_rate[timestamp] = str(item.rate)

        # tracks: list[Track] = Tracker.load_from_log(self.log_file)
        # for track in tracks:
        #     for file in track.file_stamp_list:
        #         file_dict[file] = track.result

        with tempfile.TemporaryDirectory(delete=False) as temp_dir:
            for file in os.listdir(self._folder):
                path = os.path.join(self._folder, file)
                if not file.endswith(".json"):
                    continue
                ph = PatchHistory().set_json_file(path).load_json()
                complete = ph.restore_all()
                for i, patch in enumerate(complete):
                    key = patch.label
                    rate = timestamp_rate.get(key, "???")
                    result = rate.rjust(3, "0").replace("None", "---")
                    output_file = os.path.join(temp_dir, f"{patch.label}_{result}_{file[:-5]}")
                    if i < 5 or i == len(complete) - 1:
                        output += f"  Extraindo: {output_file}\n"
                    elif i == 5:
                        output += "...\n"
                    with open(output_file, 'w', encoding='utf-8') as f:
                        f.write(patch.content)
            output += f"\nPasta com os arquivos: {temp_dir}\n"
            return output, temp_dir


    def set_files(self, files: list[Path]):
        self._files = files
        return self
    
    def set_result(self, result: str):
        self._result = result
        return self
    
    def set_percentage(self, percentage: int):
        self._result = "{}%".format(str(percentage).rjust(3, "0"))
        return self

    def get_log_full_path(self):
        return os.path.join(self._folder, Tracker.log_file)

    def set_folder(self, folder: Path):
        self._folder = folder
        return self
    
    # in format: YYYY-MM-DD HH:MM:SS
    @staticmethod
    def get_timestamp():
        return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    
    @staticmethod
    def get_timestamp_from_string(timestamp: str) -> datetime.datetime:
        try:
            return datetime.datetime.strptime(timestamp, "%Y-%m-%d_%H-%M-%S")
        except ValueError:
            raise ValueError(f"Invalid timestamp format: {timestamp}. Expected format is YYYY-MM-DD_HH-MM-SS.")

    # return timestamp of the last version of the file
    def save_file_with_timestamp_prefix(self, timestamp: str, file: Path) -> tuple[str, bool, int]:
        filename = os.path.basename(file)
        ph = PatchHistory()
        json_file = os.path.join(self._folder, f"{filename}{Tracker.extension}")
        ph.set_json_file(json_file)
        ph.load_json()

        content = Decoder.load(file)
        last_version = ph.store_version(timestamp, content)
        if last_version == timestamp:
            ph.save_json()
            return timestamp, True, len(content.splitlines())
        return last_version, False, len(content.splitlines())

    # return True if any file was changed
    def store(self) -> tuple[bool, int]:
        os.makedirs(self._folder, exist_ok=True)

        files_in_this_version: list[str] = []
        timestamp = self.get_timestamp()

        any_changes = False
        total_size = 0
        for file in self._files:
            stored, changed, size = self.save_file_with_timestamp_prefix(timestamp, file)
            total_size += size
            filename = os.path.basename(file)
            files_in_this_version.append(filename + ":" + stored)
            if changed:
                any_changes = True

        log_file = self.get_log_full_path()
        track = Track().set_timestamp(timestamp).set_file_stamp_list(files_in_this_version).set_result(self._result)
        with open(log_file, encoding="utf-8", mode="a", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(track.track_to_column())
        return any_changes, total_size
    
    @staticmethod
    def load_from_log(log_file: str) -> list[Track]:
        if not os.path.exists(log_file):
            return []
        
        tracks: list[Track] = []
        with open(log_file, encoding="utf-8") as f:
            reader = csv.reader(f)
            for row in reader:
                if len(row) < 3:
                    continue
                track = Track().column_to_track(row)
                tracks.append(track)
        return tracks

    @staticmethod
    def main():
        parser = argparse.ArgumentParser(description="Track files changes.")
        parser.add_argument("files", metavar="files", type=str, nargs="+", help="files to be tracked.")
        args = parser.parse_args()
        
        tracker = Tracker().set_folder(Path(".track")).set_files(args.files)
        tracker.store()



from __future__ import annotations
from abc import ABC, abstractmethod
import datetime as dt
from tko.logger.delta import Delta
import enum
from tko.logger.kv import KV

class LogItemBase(ABC):
    class Type(enum.Enum):
        EXEC = 'EXEC'  # Execution Free or With tests
        SELF = 'SELF'  # Self evaluation
        MOVE = 'MOVE'  # Interaction with tasks without execution 
        GAME = 'GAME'  # Game changes or setup

    key_str = "k"
    version_str = "v"
    def __init__(self, log_type: LogItemBase.Type):
        self.key = ""
        self.vers: int = 1
        self.datetime: dt.datetime = dt.datetime.fromordinal(1)
        self.timestamp: str = ""
        self.type: LogItemBase.Type = log_type

    def set_key(self, key: str):
        self.key = key
        return self
    
    def set_version(self, version: int):
        self.vers = version
        return self
    
    def get_version(self) -> int:
        return self.vers

    def get_key(self) -> str:
        return self.key

    def set_timestamp(self, str_timestamp: str, datetime: dt.datetime | None = None):
        self.timestamp = str_timestamp
        if datetime is not None:
            self.datetime = datetime
        else:
            self.datetime = Delta.decode_format(str_timestamp)
        return self

    def set_datetime(self, datetime: dt.datetime):
        self.datetime = datetime
        self.timestamp = Delta.encode_format(datetime)
        return self

    def get_timestamp(self) -> str:
        return self.timestamp

    def get_datetime(self) -> dt.datetime:
        return self.datetime

    def get_log_type(self) -> LogItemBase.Type:
        return self.type

    def __str__(self) -> str:
        return self.encode_line()

    @abstractmethod
    def identify_kv(self, kv: dict[str, str]) -> bool:
        """
        Identify key-value pairs in the dictionary and set the corresponding attributes.
        Returns True if successful, False otherwise.
        """
        pass

    def decode_line(self, parts: list[str]) -> bool:
        kv = KV.decode_args(parts[2:])
        self.set_timestamp(parts[0])
        self.type = LogItemBase.Type(parts[1])
        self.vers = int(kv.get(self.version_str, 1))
        self.key = kv.get(self.key_str, "")
        return self.identify_kv(kv)

    def encode_line(self) -> str:
        """
        Encode the log item into a line for the log file.
        Returns the encoded line as a string.
        """
        return f'{self.timestamp}, {self.type.value}, {self.version_str}:{self.vers}, {self.key_str}:{self.key}'


from tko.play.floating import FloatingABC

class FloatingManager:
    def __init__(self):
        self.input_layer: list[FloatingABC] = []

    def add_input(self, floating: FloatingABC):
        self.input_layer.append(floating)

    def draw_warnings(self):
        if len(self.input_layer) > 0 and self.input_layer[0].is_enable():
            self.input_layer[0].draw()

    def has_floating(self) -> bool:
        while len(self.input_layer) > 0 and not self.input_layer[0].is_enable():
            self.input_layer = self.input_layer[1:]
        return len(self.input_layer) > 0 and self.input_layer[0].is_enable()

    def draw(self) -> None:
        self.input_layer[0].draw()

    def process_input(self, key: int) -> int:
        return self.input_layer[0].process_input(key)



from pathlib import Path

from tko.down.drafts import Drafts
from tko.game.quest_parser import QuestParser
from tko.game.task_parser import TaskParser
from tko.game.quest import Quest
from tko.game.task import Task
from tko.settings.rep_source import RepSource
from tko.util.decoder import Decoder
import os
from icecream import ic # type: ignore

class GameBuilder:
    def __init__(self, source: RepSource):
        self.source = source
        self.ordered_quests: list[str] = [] # ordered quests keys
        self.quests: dict[str, Quest] = {}
        self.active_quest: Quest | None = None
        self.unique_keys: set[str] = set()
        self.interactive: bool = False

    def set_interactive(self, interactive: bool):
        self.interactive = interactive
        return self

    def build_from(self, language: str):
        filename: Path = self.source.get_source_readme()
        content: str = ""
        if not filename.exists():
            print(f"Aviso: fonte {filename} não encontrada no source {self.source.name}")
        else:
            content = Decoder.load(filename)
        self.__parse_file_content(content)
        self.__clear_empty_or_other_language(language)
        self.__create_requirements_pointers()
        self.__create_cross_references()
        return self


    def collect_tasks(self) -> dict[str, Task]:
        tasks: dict[str, Task] = {}

        for quest in self.quests.values():
            for task in quest.get_tasks():
                tasks[task.get_full_key()] = task
        return tasks

    def collect_quests(self) -> dict[str, Quest]:
        quests: dict[str, Quest] = {}
        for quest in self.quests.values():
            quests[quest.get_full_key()] = quest
        return quests

    def __create_requirements_pointers(self):
        quests, tasks = self.source.get_filters()
        if quests is not None or tasks is not None:
            return

        filename: Path = self.source.get_source_readme()
        quests = self.collect_quests()
        # verificar se todas as quests requeridas existem e adicionar o ponteiro
        for q in quests.values():
            for r in q.requires:
                if r in quests:
                    q.requires_ptr.append(quests[r])
                else:
                    # print(f"keys: {self.quests.keys()}")
                    print(f"Quest\n{filename}:{q.line_number}\n{str(q)}\nrequer {r} que não existe")
                    exit(1)

    @staticmethod
    # extract yaml keys in front matter format and return as dict, also return content without yaml front matter
    def extract_yaml_keys_and_file_data(content: str) -> tuple[dict[str, str], str]:
        yaml_keys: dict[str, str] = {}
        lines = content.splitlines()
        if len(lines) > 0 and lines[0].strip() == "---":
            for line in lines[1:]:
                if line.strip() == "---":
                    break
                if ":" in line:
                    key, val = line.split(":", 1)
                    yaml_keys[key.strip()] = val.strip()
                if "=" in line:
                    key, val = line.split("=", 1)
                    yaml_keys[key.strip()] = val.strip()
            content = "\n".join(lines[len(yaml_keys) + 2:])
        return yaml_keys, content

    def __parse_quest_folder(self, task_dir_path: str, tasks_with_missing_keys: list[str]):
        database_path = task_dir_path
        alias = self.source.name
        if not os.path.exists(database_path):
            os.makedirs(database_path)
        for task_dirname in sorted(os.listdir(database_path)):
            task_dir_path = os.path.join(database_path, task_dirname)
            if not os.path.isdir(task_dir_path):
                continue
            if not os.path.isfile(os.path.join(task_dir_path, "README.md")):
                continue
            # parse readme content for yaml front matter with key
            yaml_keys, _ = self.extract_yaml_keys_and_file_data(Decoder.load(os.path.join(task_dir_path, "README.md")))
            if "key" not in yaml_keys or yaml_keys["key"] in self.unique_keys:
                tasks_with_missing_keys.append(task_dir_path)
                # print(f"Readme {task_dir_path}/README.md não possui chave yaml única, será ignorado")
            else:
                self.unique_keys.add(yaml_keys["key"])
                self.create_task_from_folder(alias, yaml_keys["key"], task_dirname, task_dir_path)

    def create_task_from_folder(self, alias: str, yaml_key: str, task_dirname: str, task_dir_path: str):
        task = Task()
        title = task_dirname
        title = f"@{yaml_key} {task_dirname}"
        task.set_key(yaml_key)
        task.set_title(title)
        task.set_remote_name(alias)
        task.set_origin_folder(Path(task_dir_path))
        if self.source.is_read_only():
            task.set_workspace_folder(self.source.get_task_workspace(yaml_key))
        self.__add_task(task)

    def __parse_remote_folder(self, folder: str):
        alias = self.source.name
        if not os.path.exists(folder):
            return
        tasks_with_missing_keys: list[str] = []
        for entry in sorted(os.listdir(folder)):
            quest_folder = os.path.join(folder, entry)
            if not os.path.isdir(quest_folder):
                continue
            self.__add_quest(Quest(entry, f"{entry}").set_remote_name(alias))
            self.__parse_quest_folder(quest_folder, tasks_with_missing_keys)
        
        if self.interactive and len(tasks_with_missing_keys) > 0:
            print(f"Os seguintes diretórios de tarefas não possuem chave yaml única e foram ignorados:")
            for path in tasks_with_missing_keys:
                print(f"  {path}")
            print("Você deseja gerar chaves únicas para essas tarefas? (s/n): ", end="")
            answer = input().strip().lower()
            if answer == "s":
                max_number = Drafts.find_max_numbered_key(list(self.unique_keys))
                for path in tasks_with_missing_keys:
                    max_number += 1
                    yaml_key = Drafts.create_draft_key(max_number)
                    yaml_keys, info = self.extract_yaml_keys_and_file_data(Decoder.load(os.path.join(path, "README.md")))
                    yaml_keys["key"] = yaml_key
                    new_content = "---\n"
                    for k, v in yaml_keys.items():
                        new_content += f"{k}: {v}\n"
                    new_content += "---\n\n" + info
                    with open(os.path.join(path, "README.md"), "w", encoding="utf-8") as f:
                        f.write(new_content)
                    self.create_task_from_folder(alias, yaml_key, os.path.basename(path), path)



    def __is_autoload_quest_cmd(self, line: str) -> tuple[bool, str]:
        words = line.split(f"{self.source.AUTOLOAD_QUEST_COMMAND}=")
        if len(words) == 2:
            path = words[1].strip("-> ")
            readme_folder = os.path.dirname(self.source.get_source_readme())
            folder = os.path.join(readme_folder, path)
            return True, folder
        return False, ""
    
    def __is_autoload_cluster_cmd(self, line: str) -> tuple[bool, str]:
        words = line.split(f"{self.source.AUTOLOAD_CLUSTER_COMMAND}=")
        if len(words) == 2:
            path = words[1].strip("-> ")
            readme_folder = os.path.dirname(self.source.get_source_readme())
            folder = os.path.join(readme_folder, path)
            folder = os.path.abspath(folder)
            return True, folder
        return False, ""


    def __parse_file_content(self, content: str):
        lines = content.splitlines()
        alias = self.source.name
        filename = self.source.get_source_readme()
        for line_num, line in enumerate(lines):
            autoload, quest_folder = self.__is_autoload_quest_cmd(line)
            if autoload:
                self.__parse_quest_folder(quest_folder, [])
                continue

            autoload, remote_folder = self.__is_autoload_cluster_cmd(line)
            if autoload:
                self.__parse_remote_folder(remote_folder)
                continue

            quest = QuestParser(alias).parse_quest(filename, line, line_num + 1)
            if quest is not None:
                self.__add_quest(quest)
                continue

            tp = TaskParser(filename, alias)
            task = tp.parse_line(line, line_num + 1).check_path_try().get_task()
            if task is not None:
                if self.source.is_read_only() and not task.is_link():
                    task.set_workspace_folder(self.source.get_task_workspace(task.get_key()))
                self.__add_task(task)

    def __get_active_quest(self) -> Quest:
        if self.active_quest is None:
            qkey = "_sem_quest"
            return self.__add_quest(Quest("Sem Quest", qkey))
        return self.active_quest

    def __add_quest(self, quest: Quest) -> Quest:
        self.quests[quest.get_full_key()] = quest
        self.ordered_quests.append(quest.get_full_key())
        self.active_quest = quest
        return quest

    def __add_task(self, task: Task):
        self.__get_active_quest().add_task(task)


    def remove_empty_and_other_language_and_filtered(self, language: str, quest_filters: list[str] | None, task_filters: list[str] | None):
        # self.__quests = [q for q in self.__quests if len(q.get_tasks()) > 0]
        quests: list[Quest] = []
        for q in self.quests.values():
            if len(q.get_tasks()) == 0:
                continue
            if quest_filters is not None and len(quest_filters) > 0:
                allow = False
                for filter in quest_filters:
                    if filter in q.get_title() or filter in q.get_full_key():
                        allow = True
                        break
                if not allow:
                    continue
            if len(q.languages) == 0 or language in q.languages:
                quests.append(q)
        self.quests = {q.get_full_key(): q for q in quests}

        return self

    def __clear_empty_or_other_language(self, language: str): #call before create_cross_references
        # apagando quests vazias da lista de quests
        quest_filters, task_filters = self.source.get_filters()
        self.remove_empty_and_other_language_and_filtered(language, quest_filters, task_filters) 

    def __create_cross_references(self): #call after clear_empty
        for quest in self.quests.values():
            quest.remote_name = self.source.name
            for task in quest.get_tasks():
                task.remote_name = self.source.name
                task.quest_key = quest.get_full_key()


# from typing import override

from tko.settings.settings import Settings
from tko.enums.diff_mode import DiffMode

class ConfigParams:
    def __init__(self):
        self.side: bool = False
        self.down: bool = False
        self.images: str | None = None
        self.editor: str | None = None
        self.borders: str | None = None
        self.timeout: int | None = None

    # @override
    def __str__(self) -> str: 
        return f"side: {self.side}, down: {self.down}, timeout: {self.timeout}, editor: {self.editor}, borders: {self.borders}, images: {self.images}"

class CmdConfig:
        
    @staticmethod
    def execute(settings: Settings, param: ConfigParams):
        action = False

        if param.borders is not None:
            action = True
            settings.app.set_use_borders(param.borders == "1")
            print("Borders now is: " + str("True" if param.borders == "1" else "False"))
        if param.images is not None:
            action = True
            settings.app.set_use_images(param.images == "1")
            print("Images now is: " + str("True" if param.images == "1" else "False"))
            
        if param.side:
            action = True
            settings.app.set_diff_mode(DiffMode.SIDE)
            print("Diff mode now is: SIDE_BY_SIDE")
        if param.down:
            action = True
            settings.app.set_diff_mode(DiffMode.DOWN)
            print("Diff mode now is: UP_DOWN")
        # if param.lang:
        #     action = True
        #     settings.app._lang_default = param.lang
        #     print("Default language extension now is:", param.lang)
        # if param.ask:
        #     action = True
        #     settings.app._lang_default = ""
        #     print("Language extension will asked be always.")
        
        if param.editor:
            action = True
            settings.app.editor = param.editor
            print(f"Novo comando para abrir arquivos de código: {param.editor}")

        if param.timeout is not None:
            action = True
            settings.app.set_timeout(param.timeout)
            print(f"Novo timeout: {param.timeout}")

        if not action:
            print(str(settings))

        settings.save_settings()


from tko.logger.logger import Logger
from uniplot import plot_to_string # type: ignore
from tko.util.text import Text
from tko.logger.delta import Delta
from icecream import ic # type: ignore

class DailyGraph:
    def __init__(self, logger: Logger, width: int, height: int):
        self.logger = logger
        self.width = width
        self.height = height
        self.daily: list[float] = []
        self.accumulates: list[float] = []
        self.eixo: list[float] = []
        # self.raw_text: list[str] = []
        self.__collect()

    def __collect(self):
        days_info = self.logger.daily.resume()
        sorted_keys = sorted(days_info.keys())
        
        self.daily = [0]
        self.accumulates = [0]
        for item in sorted_keys:
            self.daily.append(days_info[item].elapsed_seconds)
            self.accumulates.append(self.accumulates[-1] + days_info[item].elapsed_seconds)
        
        for i, value in enumerate(self.daily):
            self.daily[i] = value / 3600
        for i, value in enumerate(self.accumulates):
            self.accumulates[i] = value / 3600

        self.eixo = list(range(len(self.daily)))

    # def get_collected(self) -> list[str]:
    #     output: list[str] = []
    #     for value in self.collected:
    #         output.append(f"{value:.2f}")
    #     for i in range(len(output)):
    #         output[i] = output[i].replace(".", ":").rjust(5, " ")
    #     return output

    def get_graph(self) -> list[Text]:
        # collected: list[float] = []
        # eixo: list[int] = []
        if not self.daily:
            return []

        daily: list[float] = [x for x in self.daily]
        accumulates: list[float] = [x for x in self.accumulates]

        max_daily = max(daily) if daily else 0
        max_accumulates = max(accumulates) if accumulates else 0

        for i in range(len(daily)):
            daily[i] = daily[i] / max_daily * 100 if max_daily > 0 else 0
            accumulates[i] = accumulates[i] / max_accumulates * 100 if max_accumulates > 0 else 0

        eixo: list[float] = []
        bar: list[float] = []
        day = 1
        for value in daily:
            eixo.append(day)
            bar.append(0)
            eixo.append(day)
            bar.append(value)
            eixo.append(day)
            bar.append(0)
            day += 1

        eixo = [x / 7 for x in eixo]
        self.eixo = [x / 7 for x in self.eixo]

        result: list[str] = plot_to_string(xs=[eixo, self.eixo], ys=[bar, accumulates], lines=[True, True], y_min=0,width=self.width, height=self.height, y_unit="%", x_unit="w")

        if isinstance(result, str):
            result = result.splitlines()
        lines: list[Text] = []

        lines.append(Text().addf("C", " Máximo diário: ").addf("{C}", f"{Delta.format_h_min(max_daily)} ")
                     .add(" ").addf("M", " Acumulado: ").addf("M", f"{Delta.format_h_min(max_accumulates)} ")
                     .center(self.width))
        
        for line in result:
            lines.append(Text.decode_raw(line))

        return lines



from tko.enums.identifier_type import IdentifierType

import os

class Identifier:
    def __init__(self):
        pass

    @staticmethod
    def get_type(target: str) -> IdentifierType:
        if os.path.isdir(target):
            return IdentifierType.OBI
        elif target.endswith(".md"):
            return IdentifierType.MD
        elif target.endswith(".tio"):
            return IdentifierType.TIO
        elif target.endswith(".toml"):
            return IdentifierType.TOML
        elif target.endswith(".vpl"):
            return IdentifierType.VPL
        else:
            return IdentifierType.SOLVER


import subprocess
import os
import glob


class Cases:

    @staticmethod
    def run(cases_file: str, source_readme: str, source_dir: str):
        # find all files in the directory terminatig with .tio or .vpl
        files = list(glob.iglob(source_dir + '/**', recursive=True))
        files = [f for f in files if os.path.isfile(f)]
        files = [f for f in files if f.endswith(".tio") or f.endswith(".vpl")]
        cmd = " ".join(["tko", "build", cases_file, source_readme] + files)
        subprocess.run(cmd, shell=True, stdout=subprocess.PIPE)



import curses
# from typing import override
from tko.util.text import Text
from tko.play.colors import Colors


class TextPosition:
    def __init__(self, y: int, x: int, text: Text):
        self.y = y
        self.x = x
        self.text = text

    # @override
    def __str__(self):
        return f"{self.y}:{self.x}:{self.text}"
    
class TokenPosition:
    def __init__(self, y: int, x: int, token: Text.Token):
        self.y = y
        self.x = x
        self.token = token

    # @override
    def __str__(self):
        return f"{self.y}:{self.x}:{self.token}"

class Fmt:
    __scr = None
    # Definindo constantes para as cores
    color_pairs: dict[str, int] = {}

    COLOR_MAP = {
        'k': curses.COLOR_BLACK,
        'r': curses.COLOR_RED,
        'g': curses.COLOR_GREEN,
        'y': curses.COLOR_YELLOW,
        'b': curses.COLOR_BLUE,
        'm': curses.COLOR_MAGENTA,
        'c': curses.COLOR_CYAN,
        'w': curses.COLOR_WHITE
    }
    @staticmethod
    def set_scr(scr: curses.window):
        Fmt.__scr = scr
        Fmt.init_colors()

    @staticmethod
    def init_colors():
        pair_number = 1
        curses.start_color()
        curses.use_default_colors()
        for fk, fg in Fmt.COLOR_MAP.items():
            if fk == "T":
                continue
            curses.init_pair(pair_number, fg, -1)
            Fmt.color_pairs[fk] = pair_number
            pair_number += 1

        for fk, fg in Fmt.COLOR_MAP.items():
            for bk, bg in Fmt.COLOR_MAP.items():
                curses.init_pair(pair_number, fg, bg)
                Fmt.color_pairs[fk + bk.upper()] = pair_number
                pair_number += 1

    @staticmethod
    def stroke(y: int, x: int, fmt: str, text: str):
        if Fmt.__scr is None:
            raise Exception("Fmt.__scr não foi inicializado")
        stdscr = Fmt.__scr
        italic = False
        underline = False
        reverse = False
        source_fmt = fmt
        if "X" in fmt or "x" in fmt:
            reverse = True
            fmt = fmt.replace("X", "").replace("x", "")
        if "/" in fmt:
            italic = True
            fmt = fmt.replace("/", "")
        if "_" in fmt:
            underline = True
            fmt = fmt.replace("_", "")

        fg_list = [c for c in fmt if c.islower()]
        bg_list = [c for c in fmt if c.isupper()]
        bg = "" if len(bg_list) == 0 else bg_list[0]
        fg = "" if len(fg_list) == 0 else fg_list[0]

        if bg != "" and fg == "":
            fg = "k"
        if fg == "" and bg == "":
            pair_number = -1
        else:
            try:
                pair_number = Fmt.color_pairs[fg + bg]
            except KeyError:
                # print("Cor não encontrada: " + fg + bg)
                raise(Exception("Cor não encontrada: " + source_fmt))
        if italic:
            stdscr.attron(curses.A_ITALIC)
        if underline:
            stdscr.attron(curses.A_UNDERLINE)
        if reverse:
            stdscr.attron(curses.A_REVERSE)
        # Exibir o texto com a combinação de cores escolhida
        if pair_number != -1:
            stdscr.attron(curses.color_pair(pair_number))
        try:
            stdscr.addstr(y, x, text)
        except curses.error as _e:
            lines, cols = Fmt.get_size()
            if y == lines - 1:
                if x + len(text) <= cols:
                    pass
        except ValueError as _e:
            pass
            # lines, cols = stdscr.getmaxyx()
            # stdscr.addstr(10, 10, f"y:{y}, x:{x}, fmt:{fmt}, len:{len(text)} lines:{lines}, cols:{cols}")
            # stdscr.addstr(1, 0, text)
            # raise Exception(f"y:{y}, x:{x}, fmt:{fmt}, len:{len(text)} lines:{lines}, cols:{cols}\n{text}")
        if pair_number != -1:
            stdscr.attroff(curses.color_pair(pair_number))
        if italic:
            stdscr.attroff(curses.A_ITALIC)
        if underline:
            stdscr.attroff(curses.A_UNDERLINE)
        if reverse:
            stdscr.attroff(curses.A_REVERSE)

    # break in lines and cut everything that is out of the box
    @staticmethod
    def cut_box(y: int, x: int, box_y: int, box_x: int, sentence: Text) -> list[TextPosition]:
        lines: list[Text] = sentence.split("\n")
        output: list[TextPosition] = []
        for line in lines:
            px = x
            if y < 0:
                y += 1
                continue
            if y >= box_y:
                break
            if px < 0:
                line.data = line.data[-px:]
                px = 0
            if px + len(line) > box_x:
                line.data = line.data[:box_x - px]
            output.append(TextPosition(y, px, line))
            y += 1
        return output

    @staticmethod
    def split_in_tokens(text_lines: list[TextPosition]) -> list[TokenPosition]:
        output: list[TokenPosition] = []
        for text_pos in text_lines:
            y = text_pos.y
            x = text_pos.x
            line = text_pos.text
            for token in line.resume():
                output.append(TokenPosition(y, x, token))
                x += len(token)
        return output

    @staticmethod
    def write(y: int, x: int, sentence: Text | str):
        if isinstance(sentence, str):
            data: Text = Text().add(sentence)
        else:
            data = sentence

        lines, cols = Fmt.get_size()
        text_lines = Fmt.cut_box(y, x, lines, cols, data)
        token_list = Fmt.split_in_tokens(text_lines)
        for token_pos in token_list:
            y = token_pos.y
            x = token_pos.x
            token = token_pos.token
            # print(f"y:{y}, x:{x}, fmt:{token.fmt}, len:{len(token.text)}")
            Fmt.stroke(y, x, token.fmt, token.text)

    # @staticmethod
    # def get_user_input(stdscr, prompt: str) -> str:
    #     lines, cols = stdscr.getmaxyx()
    #     curses.echo()  # Ativa a exibição dos caracteres digitados
    #     curses.curs_set(1)  # Ativa o cursor
    #     stdscr.addstr(0, 0, cols * " ")
    #     stdscr.addstr(0, 0, prompt)
    #     stdscr.refresh()
    #     input_str = stdscr.getstr(0, len(prompt), 20).decode('utf-8')  # Captura o input do usuário
    #     curses.noecho()  # Desativa a exibição dos caracteres digitados
    #     curses.curs_set(0)
    #     return input_str

    @staticmethod
    def get_percent(value: int, pad: int = 0) -> Text:
        colors = Colors()
        text = f"{str(value)}%".rjust(pad)
        if value == 100:
            return Text().addf(colors.mark_complete, "100%")
        if value >= 70:
            return Text().addf(colors.mark_required, text)
        if value == 0:
            return Text().addf(colors.mark_nothing, text)
        return Text().addf(colors.mark_started, text)
    
    @staticmethod
    def get_screen() -> curses.window:
        if Fmt.__scr is None:
            raise Exception("Fmt.__scr não foi inicializado")
        return Fmt.__scr

    @staticmethod
    def getch():
        return Fmt.get_screen().getch()

    @staticmethod
    def clear():
        Fmt.get_screen().erase()

    @staticmethod
    def refresh():
        Fmt.get_screen().refresh()

    @staticmethod
    def get_size() -> tuple[int, int]:
        return Fmt.get_screen().getmaxyx()
        

def test_fmt(scr: curses.window):
    curses.curs_set(0)  # Esconde o cursor
    Fmt.init_colors()  # Inicializa as cores
    Fmt.set_scr(scr)  # Define o scr como global

    output = Text.format("..")
    for i in range(60):
        output.addf("r", str(i) + " ")
    for i in range(15):
        Fmt.write(i - 2, -1, output)
    scr.getch()



from __future__ import annotations

import argparse
from tko.util.text import Text
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
import subprocess
from pathlib import Path


class Pull:

    @staticmethod
    def run_git_command(directory: str, command_args: list[str]) -> tuple[str, str, int]:
        try:
            result = subprocess.run(
                ["git"] + command_args,
                cwd=directory,
                capture_output=True,
                text=True,
                timeout=120,
            )
            return result.stdout, result.stderr, result.returncode
        except FileNotFoundError:
            return "", "git not found", 127
        except Exception as e:
            return "", f"unexpected error: {e}", 1

    @staticmethod
    def git_ok(directory: str, *args: str) -> tuple[bool, str]:
        stdout, stderr, code = Pull.run_git_command(directory, list(args))
        return code == 0, stderr or stdout

    @staticmethod
    def get_default_branch(directory: str) -> str:
        stdout, _, code = Pull.run_git_command(
            directory,
            ["symbolic-ref", "refs/remotes/origin/HEAD"]
        )
        if code == 0 and stdout:
            return stdout.strip().split("/")[-1]
        return "main"

    @staticmethod
    def is_up_to_date(directory: str, branch: str) -> bool:
        local, _, _ = Pull.run_git_command(directory, ["rev-parse", "HEAD"])
        remote, _, _ = Pull.run_git_command(directory, ["ls-remote", "origin", branch])

        if not remote:
            return False

        remote_hash = remote.split()[0]
        return local.strip() == remote_hash

    @staticmethod
    def pull(directory: str) -> Text:
        repo_path = Path(directory)

        if not repo_path.is_dir() or not (repo_path / ".git").exists():
            return Text()

        output = Text("g") + directory + " "

        branch = Pull.get_default_branch(directory)

        # 1. Skip se já atualizado
        if Pull.is_up_to_date(directory, branch):
            return output + Text("c") + "Up-to-date"

        # 2. Fetch otimizado
        output += Text("y") + "Fetch "
        ok, msg = Pull.git_ok(
            directory,
            "fetch",
            "--prune",
            "--filter=blob:none",
            "--tags",
            "origin",
            branch
        )
        if not ok:
            return output + "\nFetch failed: " + msg

        # 3. Aplicar atualização
        output += Text("y") + "Update "
        ok, msg = Pull.git_ok(directory, "reset", "--hard", "FETCH_HEAD")

        if not ok:
            # fallback raro (repo zoado)
            output += Text("r") + "Fallback "
            Pull.git_ok(directory, "clean", "-fd")
            ok, msg = Pull.git_ok(directory, "reset", "--hard", f"origin/{branch}")
            if not ok:
                return output + "\nReset failed: " + msg

        return output + "\n"

    @staticmethod
    def pull_all_parallel(repo_list: list[Path], max_workers: int = 10):
        print(f"\n{Text('y')}Pull de {len(repo_list)} repositórios ({max_workers} threads){Text('e')}")
        start = time.time()

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(Pull.pull, str(repo)): repo for repo in repo_list}

            for future in as_completed(futures):
                repo = futures[future]
                try:
                    output = future.result()
                    if output.get_str():
                        print(output)
                except Exception as exc:
                    print(f"{Text('r')}Erro em {repo}: {exc}{Text('e')}")

        elapsed = time.time() - start
        print(f"\n{Text('y')}Finalizado em {elapsed:.2f}s{Text('e')}")


def pull_all_parallel_main(args: argparse.Namespace) -> None:
    path_list = [Path(p) for p in args.path]
    n_threads = args.threads if args.threads else 10
    Pull.pull_all_parallel(path_list, n_threads)


from __future__ import annotations
from typing import Any
from tko.play.keys import GuiKeys

class Flag:
    def __init__(self):
        self._name: str = ""
        self._text: str = ""  # description
        self._msgs: list[str] = []
        self._char: str = ""
        self._values: list[str] = ["0", "1"]
        self._index: int = 0
        self._location: str = ""
        self._bool = True  # many options
        self._autoload = True

    def set_autoload(self, value: bool):
        self._autoload = value
        return self
    
    def get_autoload(self) -> bool:
        return self._autoload

    def set_name(self, _name: str):
        self._name = _name
        return self

    def set_description(self, _text: str):
        self._text = _text
        return self
    
    def get_msgs(self) -> list[str]:
        return self._msgs
    
    def set_msgs(self, msgs: list[str]):
        self._msgs = msgs
        return self

    def set_keycode(self, _key: str):
        self._char = _key
        return self

    def set_values(self, _values: list[str]):
        self._values = _values
        return self

    def index(self, _index: int):
        self._index = _index
        return self

    def get_values(self) -> list[str]:
        return self._values

    def toggle(self):
        self._index = (self._index + 1) % len(self._values)
        return None

    def get_value(self) -> str:
        return self._values[self._index % len(self._values)]
    
    def set_value(self, value: Any):
        for i, v in enumerate(self._values):
            if v == value:
                self._index = i
                break

    def get_name(self) -> str:
        return self._name

    def get_description(self) -> str:
        return self._text

    def get_keycode(self) -> str:
        return self._char

    def get_index(self) -> int:
        return self._index
    
    def is_true(self) -> bool:
        return self.get_value() != "0"
    
    def __bool__(self) -> bool:
        return self.get_value() == "1"
    

class Flags:
    panel_help = "0"
    panel_graph = "1"
    panel_logs = "2"
    panel_skills = "3"

    task_exec_view = "0"
    task_time_view = "1"
    
    inbox_only = "0"
    inbox_all = "1"

    panel = (Flag().set_name("Painel").set_description("Mostra o Painel de Informações")
                    .set_values([panel_help, panel_graph, panel_logs, panel_skills])
                    .set_msgs(["Desabilitar painel de informações", "Gráfico de tarefas", "Mostrar logs", "Mostrar trilhas de habilidades"]))
    inbox = (Flag().set_name("Tópicos").set_description("Habilitas todas as missões e tarefas")
                    .set_values([inbox_only, inbox_all])
                    .set_msgs(["Mostrar inbox", "Habilitar todas as tarefas"]))
    show_panel = (Flag().set_name("ShowPanel").set_description("Mostra o painel esquerdo")
                    .set_values(["0", "1"])
                    .set_msgs(["Desabilitar painel lateral", "Mostrar painel lateral"]))
    tasks = (Flag().set_name("Tarefas").set_description("Mostra as atividades concluídas")
                    .set_values(["0", "1", "2"])
                    .set_msgs(["Mostrar todas as tarefas", "Ocultar tarefas com 100%", "Ocultar tarefas com >70%"]))
    task_graph_mode = (Flag().set_name("Task Graph").set_description("Mostra o Gráfico de Tarefas")
                    .set_values([task_exec_view, task_time_view])
                    .set_msgs(["Gráfico de tarefas por execuções", "Gráfico de tarefas por tempo"]))
    show_time = (Flag().set_name("Tempo").set_description("Mostra o tempo utilizado para completar as tarefas")
                    .set_values(["1", "0"])
                    .set_keycode(GuiKeys.show_duration)
                    .set_msgs(["Mostrar tempo gasto nas tarefas", "Ocultar tempo gasto nas tarefas"]))
class FlagsMan:
    def __init__(self, data: dict[str, int]):
        self.flags: dict[str, Flag] = {}

        for varname, flag in Flags.__dict__.items():
            if isinstance(flag, Flag):
                self.flags[varname] = flag

        for key, _index in data.items():
            if key in self.flags:
                self.flags[key].index(_index)

    def get_data(self) -> dict[str, int]:
        data: dict[str, int] = {}
        for name, flag in self.flags.items():
            if len(flag.get_values()) > 1:
                data[name] = flag.get_index()
        return data



from __future__ import annotations
from tko.logger.old_log_loader import OldLogLoader
from tko.logger.log_item_base import LogItemBase
from tko.logger.log_item_self import LogItemSelf
from tko.logger.log_item_exec import LogItemExec
from tko.logger.log_item_move import LogItemMove
from tko.logger.delta import Delta
from tko.settings.rep_paths import RepPaths
import datetime as dt

from tko.util.decoder import Decoder
from typing import Callable
from pathlib import Path

class LogHistory:

    def __init__(self, rep_folder: Path, listeners: list[Callable[[LogItemBase, bool], None]] | None = None):
        if listeners is None:
            listeners = []
        self.paths = RepPaths(rep_folder)
        self.log_folder: Path = self.paths.get_log_folder()
        self.listeners: list[Callable[[LogItemBase, bool], None]] = listeners
        self.entries: dict[dt.datetime, LogItemBase] = {}
        self.entries.update(self.__load_old_log())
        self.entries.update(self.__load_daily_log_folder())
        # avoid duplicated entries
        sorted_entries = sorted(self.entries.items(), key=lambda x: x[0])
        
        for _, item in sorted_entries:
            for listener in self.listeners:
                listener(item, False)

    def __load_old_log(self) -> dict[dt.datetime, LogItemBase]:
        self.old_log_file = self.paths.get_old_history_file()
        loader = OldLogLoader(self.paths.get_workspace_dir())
        return loader.base_dict

    def get_entries(self) -> dict[dt.datetime, LogItemBase]:
        return self.entries

    def get_log_folder(self) -> Path:
        return self.log_folder


    @staticmethod
    def log_file_for_day(folder: Path, datetime: dt.datetime) -> Path:
        date_str = datetime.strftime("%Y-%m-%d")
        folder.mkdir(parents=True, exist_ok=True)
        return folder / f"{date_str}.log"

    def append_new_action(self, item_base: LogItemBase) -> LogItemBase:
        now_str, now_dt = Delta.now()
        item_base.set_timestamp(now_str, now_dt)
        self.entries[now_dt] = item_base
        for listener in self.listeners:
            listener(item_base, True)

        log_folder = self.get_log_folder()
        log_file = LogHistory.log_file_for_day(log_folder, item_base.get_datetime())
        with open(log_file, 'a', encoding="utf-8", newline='') as file:
            file.write(f'{item_base.encode_line()}\n')
        return item_base

    def __load_daily_log_folder(self) -> dict[dt.datetime, LogItemBase]:
        log_folder = self.paths.get_log_folder()
        if not log_folder.exists():
            return {}
        if not log_folder.is_dir():
            raise ValueError(f"Log folder '{log_folder}' is not a directory.")
        files = log_folder.iterdir()

        files_path = [f for f in files if f.suffix == '.log']
        if not files_path:
            return {}
        # begin with the older file
        files_path.sort()
        entries: dict[dt.datetime, LogItemBase] = {}
        for file in files_path:
            content = Decoder.load(file)
            for line in content.splitlines():
                item: LogItemBase | None = LogHistory.decode_line(line)
                if item is not None:
                    entries[item.get_datetime()] = item
        return entries
    
    @staticmethod
    def decode_line(line: str) -> LogItemBase | None:
        parts = line.strip().split(", ")
        if len(parts) < 2:
            return None
        item_type = LogItemBase.Type(parts[1])
        if item_type == LogItemBase.Type.MOVE:
            item = LogItemMove()
            if item.decode_line(parts):
                return item
        elif item_type == LogItemBase.Type.EXEC:
            item = LogItemExec()
            if item.decode_line(parts):
                return item
        elif item_type == LogItemBase.Type.SELF:
            item = LogItemSelf()
            if item.decode_line(parts):
                return item
        return None



from __future__ import annotations
from tko.logger.task_resume import TaskResume
from typing import Any

default_value: int = 1
default_leet: bool = True
default_opt: bool = False

class Game:
    key_str: str = "key"
    value_str: str = "value"
    opt_str: str = "opt"
    tasks_str: str = "tasks"
    quests_str: str = "quests"
    leet_str: str = "leet"

    class Task:
        def __init__(self, key: str = "", value: int = default_value, is_leet: bool = default_leet, opt: bool = default_opt):
            self.key = key
            self.value: int = value
            self.opt: bool = opt
            self.leet: bool = is_leet
        
        def to_dict(self) -> dict[str, Any]:
            output: dict[str, Any] = {}
            output[Game.key_str] = self.key
            if self.value != default_value:
                output[Game.value_str] = self.value
            if self.leet != default_leet:
                output[Game.leet_str] = self.leet
            if self.opt != default_opt:
                output[Game.opt_str] = self.opt
            return output
        
        def load_from_dict(self, json_data: dict[str, Any]):
            self.key = json_data.get(Game.key_str, self.key)
            self.value = json_data.get(Game.value_str, self.value)
            self.leet = json_data.get(Game.leet_str, self.leet)
            self.opt = json_data.get(Game.opt_str, self.opt)
            return self

        def __str__(self):
            return f"{self.key}, {Game.value_str}:{self.value}, leet:{self.leet}, {Game.opt_str}:{self.opt}"

    class Quest:
        def __init__(self, key: str):
            self.key = key
            self.tasks: list[Game.Task] = []

        def to_dict(self) -> dict[str, Any]:
            output: dict[str, Any] = {
                Game.key_str: self.key,
                Game.value_str: self.value,
                Game.tasks_str: [task.to_dict() for task in self.tasks]
            }
            return output

        def load_from_dict(self, json_data: dict[str, Any]):
            self.key = json_data.get(Game.key_str, self.key)
            self.value = json_data.get(Game.value_str, self.value)
            tasks_data = json_data.get(Game.tasks_str, [])
            for task in tasks_data:
                collected_task = Game.Task().load_from_dict(task)
                self.tasks.append(collected_task)
            return self

        def __str__(self):
            return f'{self.key}, {Game.value_str}:{self.value}\n' + "\n".join([f"\t{str(t)}" for t in self.tasks])

class Collected:
    quests_str: str = "quests"
    resume_str: str = "resume"
    graph_str: str = "graph"
    log_str: str = "log"

    def __init__(self):
        self.resume: dict[str, TaskResume] = {}
        self.graph: str = ""
        self.log: list[str] = []
        self.quests: list[Game.Quest] = []

    def load_from_dict(self, json_data: dict[str, Any]):
        if not Collected.resume_str in json_data:
            print("No resume data found in the JSON.")
            return self

        task_resume = json_data.get(Collected.resume_str, self.resume)
        for key, value in task_resume.items():
            collected_resume = TaskResume(key)
            collected_resume.from_dict(value)
            self.resume[key] = collected_resume
        self.graph = json_data.get(Collected.graph_str, self.graph)
        self.log = json_data.get(Collected.log_str, self.log)
        quest_data = json_data.get(Collected.quests_str, [])
        for quest in quest_data:
            game_quest = Game.Quest(quest.get(Game.key_str, ""))
            game_quest.load_from_dict(quest)
            self.quests.append(game_quest)
        return self

    def to_dict(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        resume_dict: dict[str, Any] = {}
        for key, value in self.resume.items():
            resume_dict[key] = value.to_dict()
        output[Collected.resume_str] = resume_dict
        output[Collected.graph_str] = self.graph
        output[Collected.log_str] = self.log
        output[Collected.quests_str] = [quest.to_dict() for quest in self.quests]
        return output



import yaml # type: ignore
from tko.logger.log_item_base import LogItemBase
from tko.logger.log_sort import LogSort
from tko.logger.delta import Delta
        
class DayInfo:
    def __init__(self, day:str = "", week_day: str = "", elapsed_seconds: int = 0):
        self.day = day
        self.week_day = week_day
        self.elapsed_seconds = elapsed_seconds

class DailyListener:
    def __init__(self):
        self.day_actions: dict[str, LogSort] = {}
        self.log_file: str | None = None

    def handle_entry_incoming(self, item: LogItemBase, new_entry: bool = False):
        day = item.datetime.strftime('%Y-%m-%d')

        if day not in self.day_actions:
            self.day_actions[day] = LogSort()

        mode = Delta.Mode(Delta.Mode.Action.with_time_threshold, minutes_limit=60)
        self.day_actions[day].add_item(mode, item)
        # delta, _ = self.day_actions[day].base_list[-1]

        # print(f"Added item for day {day}: {item} with result {delta}") 

    def resume(self):
        days = sorted(self.day_actions.keys())
        if days:
            day = days[0]
            last = days[-1]
            while True:
                if day not in self.day_actions:
                    self.day_actions[day] = LogSort()
                if day == last:
                    break
                day = Delta.next_day(day)

        output: dict[str, DayInfo] = {}
        for day, log_sort in self.day_actions.items():
            week_day = Delta.week_day(day)
            base_list = log_sort.base_list
            if base_list:
                delta, _ = base_list[-1]
                elapsed = int(delta.accumulated.total_seconds())
            else:
                elapsed = 0
            output[day] = DayInfo(day = day, week_day=week_day, elapsed_seconds=elapsed)
        return output
    
    def save_yaml(self):
        if self.log_file is None:
            return
        with open(self.log_file, 'w') as file:
            yaml.dump(self.day_actions, file)



from tko.game.game import Game
from tko.play.fmt import Fmt
from tko.play.language_setter import LanguageSetter
from tko.settings.settings import Settings
from tko.settings.repository import Repository
from tko.play.keys import GuiKeys
from tko.play.images import opening
from tko.play.floating_calibrate import FloatingCalibrate
from tko.play.input_manager import InputManager
from tko.play.play_palette import PlayPalette
from tko.play.floating import Floating
from tko.play.floating_manager import FloatingManager
from tko.play.flags import Flags, FlagsMan
from tko.play.tasktree import TaskTree
from tko.play.gui import Gui
from tko.play.play_actions import PlayActions
from tko.play.flag_functors import FlagFunctor

import curses

class Play:
    def __init__(self, settings: Settings, rep: Repository):
        self.settings = settings
        self.app = settings.app
        self.rep = rep
        self.game: Game = rep.game
        self.exit = False

        self.flagsman = FlagsMan(self.rep.data.flags)
        self.fman = FloatingManager()
        self.tree = TaskTree(self.settings, rep, self.fman)
        self.gui = Gui(tree=self.tree, flagsman=self.flagsman, fman=self.fman)
        self.actions = PlayActions(self.gui)
        self.play_palette = PlayPalette(self.actions)
        # self.fman.add_input( Floating().set_content(opening['yoda'].splitlines()).set_warning() )

    def display_need_update(self):
        self.gui.set_need_update()

    def save_to_json(self):
        self.tree.save_on_rep()
        self.rep.data.flags = self.flagsman.get_data()
        self.settings.save_settings()
        self.rep.save_config()

    def send_quit_msg(self):
        def set_exit():
            self.exit = True

        self.fman.add_input(
            Floating().set_content(opening['cool'].splitlines()).set_exit_fn(set_exit).set_exit_key('q').set_warning()
        )

    def move_up(self):
        self.tree.move_up()
        self.gui.xray_offset = 0

    def move_down(self):
        self.tree.move_down()
        self.gui.xray_offset = 0

    def page_up(self):
        if Flags.panel.get_value() == Flags.panel_logs:
            self.gui.xray_offset -= 5
        elif Flags.panel.get_value() == Flags.panel_graph:
            Flags.task_graph_mode.set_value(Flags.task_exec_view)
    
    def page_down(self):
        if Flags.panel.get_value() == Flags.panel_logs:
            self.gui.xray_offset += 5
        elif Flags.panel.get_value() == Flags.panel_graph:
            Flags.task_graph_mode.set_value(Flags.task_time_view)


    def move_left(self):
        self.gui.xray_offset = 0
        self.tree.arrow_left()

    def move_right(self):
        self.gui.xray_offset = 0
        self.tree.arrow_right()

    def activate(self):
        self.gui.xray_offset = 0
        return self.actions.select_task()

    def make_callback(self) -> InputManager:
        cman = InputManager()

        cman.add_str(GuiKeys.key_quit, self.send_quit_msg)
        cman.add_int(curses.KEY_EXIT, self.send_quit_msg)
        cman.add_int(curses.KEY_UP, self.move_up)
        cman.add_int(curses.KEY_DOWN, self.move_down)
        cman.add_int(curses.KEY_LEFT, self.move_left)
        cman.add_int(curses.KEY_RIGHT, self.move_right)
        cman.add_str(GuiKeys.reload_game, self.actions.reload_game)

        cman.add_int(curses.KEY_NPAGE, self.page_down)
        cman.add_int(curses.KEY_PPAGE, self.page_up)

        cman.add_str(GuiKeys.calibrate, lambda: self.fman.add_input(FloatingCalibrate(self.settings)))
        cman.add_str(GuiKeys.activate, self.activate) # type: ignore
        # cman.add_str(GuiKeys.open_url, self.actions.open_link)
        cman.add_str(GuiKeys.down_task, self.actions.down_remote_task)
        cman.add_str(GuiKeys.borders, self.app.toggle_borders)
        cman.add_str(GuiKeys.images, self.app.toggle_images)
        cman.add_str(GuiKeys.set_lang_drafts, self.gui.language.set_language)
        cman.add_str(GuiKeys.create_draft, self.actions.create_draft)
        cman.add_int(curses.KEY_DC, self.actions.delete_folder_ask)
        cman.add_str(GuiKeys.delete_folder, self.actions.delete_folder_ask)

        cman.add_str(GuiKeys.self_evaluate, self.actions.self_evaluate)
        cman.add_str(GuiKeys.inbox, lambda: Flags.inbox.set_value(Flags.inbox_only))
        cman.add_str(GuiKeys.all_tasks, lambda: Flags.inbox.set_value(Flags.inbox_all))
        cman.add_str(GuiKeys.toggle_right, lambda: Flags.show_panel.toggle())

        cman.add_str(GuiKeys.panel_help, self.open_help)
        cman.add_str(GuiKeys.panel_graph, lambda: Flags.panel.set_value(Flags.panel_graph))
        cman.add_str(GuiKeys.panel_logs, lambda: Flags.panel.set_value(Flags.panel_logs))
        cman.add_str(GuiKeys.panel_skills, lambda: Flags.panel.set_value(Flags.panel_skills))

        cman.add_str(GuiKeys.unfold_patch, self.actions.open_versions)
        
        for flag in self.flagsman.flags.values():
            if flag.get_autoload():
                cman.add_str(flag.get_keycode(), FlagFunctor(flag, self.fman))

        cman.add_str(GuiKeys.search, self.gui.search.toggle_search)
        cman.add_str(GuiKeys.palette, self.play_palette.command_pallete)

        return cman
        

    def open_help(self):
        Flags.panel.set_value(Flags.panel_help)
        Flags.show_panel.set_value("1")

    def send_char_not_found(self, key: int):
        exclude_str = [ord(v) for v in [" ", "\n"]]
        exclude_int = [ -1, curses.KEY_EXIT, curses.KEY_UP, curses.KEY_DOWN, curses.KEY_LEFT, curses.KEY_RIGHT, curses.KEY_BACKSPACE]
        if key in exclude_int + exclude_str:
            return
        if key < 0 or key > 255:
            return
        self.fman.add_input( Floating().bottom().right().set_error().put_text(f"Tecla char:{chr(key)}, code:{key}, não reconhecida") )


    def main(self, scr: curses.window):
        InputManager.fix_esc_delay()
        curses.curs_set(0)  # Esconde o cursor
        Fmt.init_colors()  # Inicializa as cores
        Fmt.set_scr(scr)  # Define o scr como global

        while True:
            self.tree.update_tree(admin_mode=Flags.inbox.get_value() == Flags.inbox_all or self.gui.search.search_mode)
            self.fman.draw_warnings()
            cman = self.make_callback()
            self.gui.show_items()

            if self.fman.has_floating():
                self.fman.draw()

            # o input tem que ser depois do draw para mostrar o floating
            value = InputManager.get_and_remap_keys(scr, self.app)
            
            if self.fman.has_floating():
                # if consumed, value becomes -1
                value = self.fman.process_input(value)
                
            if self.exit:
                break

            if self.gui.search.search_mode:
                self.gui.search.process_search(value)
            elif cman.has_int_key(value):
                callback = cman.exec_call(value)
                if callback is not None:
                    return callback
            else:
                self.send_char_not_found(value)

            self.tree.reload_sentences()
            self.save_to_json()
            
    def play(self):
        LanguageSetter.check_lang_in_text_mode(self.rep)

        while True:
            output = curses.wrapper(self.main)
            if output is None:
                return
            else:
                output()
